"""Tools for working with YAML files.

Usage:
from praisonaiagents.tools import yaml_tools
data = yaml_tools.read_yaml("config.yaml")

or
from praisonaiagents.tools import read_yaml, write_yaml, merge_yaml
data = read_yaml("config.yaml")
"""

import logging
from typing import List, Dict, Union, Optional, Any
from importlib import util
import os
from copy import deepcopy

class YAMLTools:
    """Tools for working with YAML files."""
    
    def __init__(self):
        """Initialize YAMLTools."""
        self._check_dependencies()
        
    def _check_dependencies(self):
        """Check if required packages are installed."""
        missing = []
        for package, module in [('pyyaml', 'yaml'), ('jsonschema', 'jsonschema')]:
            if util.find_spec(module) is None:
                missing.append(package)
        
        if missing:
            raise ImportError(
                f"Required packages not available. Please install: {', '.join(missing)}\n"
                f"Run: pip install {' '.join(missing)}"
            )

    def read_yaml(
        self,
        filepath: str,
        safe_load: bool = True,
        encoding: str = 'utf-8'
    ) -> Optional[Union[Dict[str, Any], List[Any]]]:
        """Read a YAML file."""
        try:
            if util.find_spec('yaml') is None:
                error_msg = "pyyaml package is not available. Please install it using: pip install pyyaml"
                logging.error(error_msg)
                return None

            # Import yaml only when needed
            import yaml
            
            with open(filepath, 'r', encoding=encoding) as f:
                if safe_load:
                    return yaml.safe_load(f)
                return yaml.load(f, Loader=yaml.FullLoader)
        except Exception as e:
            error_msg = f"Error reading YAML file {filepath}: {str(e)}"
            logging.error(error_msg)
            return None

    def write_yaml(
        self,
        data: Union[Dict[str, Any], List[Any]],
        filepath: str,
        default_flow_style: bool = False,
        encoding: str = 'utf-8',
        sort_keys: bool = False,
        allow_unicode: bool = True
    ) -> bool:
        """Write data to YAML file."""
        try:
            if util.find_spec('yaml') is None:
                error_msg = "pyyaml package is not available. Please install it using: pip install pyyaml"
                logging.error(error_msg)
                return False

            # Import yaml only when needed
            import yaml
            
            with open(filepath, 'w', encoding=encoding) as f:
                yaml.dump(
                    data,
                    f,
                    default_flow_style=default_flow_style,
                    sort_keys=sort_keys,
                    allow_unicode=allow_unicode
                )
            return True
        except Exception as e:
            error_msg = f"Error writing YAML file {filepath}: {str(e)}"
            logging.error(error_msg)
            return False

    def merge_yaml(
        self,
        files: List[str],
        output_file: Optional[str] = None,
        strategy: str = 'deep'
    ) -> Union[Dict[str, Any], List[Any], None]:
        """Merge multiple YAML files."""
        try:
            if util.find_spec('yaml') is None:
                error_msg = "pyyaml package is not available. Please install it using: pip install pyyaml"
                logging.error(error_msg)
                return None

            def deep_merge(source: dict, destination: dict) -> dict:
                """Deep merge two dictionaries."""
                for key, value in source.items():
                    if key in destination:
                        if isinstance(value, dict) and isinstance(destination[key], dict):
                            destination[key] = deep_merge(value, destination[key])
                        elif isinstance(value, list) and isinstance(destination[key], list):
                            destination[key].extend(value)
                        else:
                            destination[key] = value
                    else:
                        destination[key] = value
                return destination
            
            result = {}
            for file in files:
                data = self.read_yaml(file)
                if data is None:
                    continue
                
                if strategy == 'deep':
                    if isinstance(data, dict):
                        result = deep_merge(data, result)
                    elif isinstance(data, list):
                        if not result:
                            result = []
                        result.extend(data)
                else:  # shallow
                    if isinstance(data, dict):
                        result.update(data)
                    elif isinstance(data, list):
                        if not result:
                            result = []
                        result.extend(data)
            
            if output_file:
                self.write_yaml(result, output_file)
            
            return result
        except Exception as e:
            error_msg = f"Error merging YAML files: {str(e)}"
            logging.error(error_msg)
            return None

    def validate_yaml(
        self,
        data: Union[Dict[str, Any], List[Any], str],
        schema: Dict[str, Any]
    ) -> bool:
        """Validate YAML data against a schema."""
        try:
            if util.find_spec('jsonschema') is None:
                error_msg = "jsonschema package is not available. Please install it using: pip install jsonschema"
                logging.error(error_msg)
                return False

            # Import jsonschema only when needed
            from jsonschema import validate, ValidationError
            
            # Load data if file path
            if isinstance(data, str):
                data = self.read_yaml(data)
                if data is None:
                    return False
            
            try:
                validate(instance=data, schema=schema)
                return True
            except ValidationError as e:
                logging.error(f"YAML validation error: {str(e)}")
                return False
        except Exception as e:
            error_msg = f"Error validating YAML: {str(e)}"
            logging.error(error_msg)
            return False

    def analyze_yaml(
        self,
        data: Union[Dict[str, Any], List[Any], str]
    ) -> Optional[Dict[str, Any]]:
        """Analyze YAML data structure."""
        try:
            # Load data if file path
            if isinstance(data, str):
                data = self.read_yaml(data)
                if data is None:
                    return None
            
            def analyze_value(value: Any) -> Dict[str, Any]:
                """Analyze a single value."""
                result = {
                    'type': type(value).__name__,
                    'size': len(value) if hasattr(value, '__len__') else 1
                }
                
                if isinstance(value, dict):
                    result['keys'] = list(value.keys())
                    result['nested_types'] = {
                        k: type(v).__name__
                        for k, v in value.items()
                    }
                elif isinstance(value, list):
                    result['element_types'] = list(set(
                        type(x).__name__ for x in value
                    ))
                
                return result
            
            result = {
                'structure': analyze_value(data),
                'stats': {
                    'total_keys': sum(1 for _ in self._walk_dict(data))
                    if isinstance(data, dict) else len(data)
                }
            }
            
            return result
        except Exception as e:
            error_msg = f"Error analyzing YAML: {str(e)}"
            logging.error(error_msg)
            return None

    def _walk_dict(self, d: Dict[str, Any]) -> Any:
        """Walk through nested dictionary."""
        for k, v in d.items():
            yield k
            if isinstance(v, dict):
                yield from self._walk_dict(v)

    def transform_yaml(
        self,
        data: Union[Dict[str, Any], List[Any], str],
        operations: List[Dict[str, Any]]
    ) -> Optional[Union[Dict[str, Any], List[Any]]]:
        """Transform YAML data using specified operations."""
        try:
            # Load data if file path
            if isinstance(data, str):
                data = self.read_yaml(data)
                if data is None:
                    return None
            
            result = deepcopy(data)
            
            for op in operations:
                op_type = op.get('type')
                path = op.get('path', '')
                value = op.get('value')
                
                if not op_type:
                    continue
                
                # Split path into parts
                parts = [p for p in path.split('/') if p]
                
                # Get reference to target location
                target = result
                parent = None
                last_key = None
                
                for part in parts[:-1]:
                    if isinstance(target, dict):
                        if part not in target:
                            target[part] = {}
                        parent = target
                        target = target[part]
                        last_key = part
                    elif isinstance(target, list):
                        idx = int(part)
                        while len(target) <= idx:
                            target.append({})
                        parent = target
                        target = target[idx]
                        last_key = idx
                
                if parts:
                    last_part = parts[-1]
                    if isinstance(target, dict):
                        if op_type == 'set':
                            target[last_part] = value
                        elif op_type == 'delete':
                            target.pop(last_part, None)
                        elif op_type == 'append' and isinstance(target.get(last_part), list):
                            target[last_part].append(value)
                    elif isinstance(target, list):
                        idx = int(last_part)
                        if op_type == 'set':
                            while len(target) <= idx:
                                target.append(None)
                            target[idx] = value
                        elif op_type == 'delete' and idx < len(target):
                            del target[idx]
                        elif op_type == 'append':
                            target.append(value)
                else:
                    if op_type == 'set':
                        result = value
                    elif op_type == 'append' and isinstance(result, list):
                        result.append(value)
            
            return result
        except Exception as e:
            error_msg = f"Error transforming YAML: {str(e)}"
            logging.error(error_msg)
            return None

# Create instance for direct function access
_yaml_tools = YAMLTools()
read_yaml = _yaml_tools.read_yaml
write_yaml = _yaml_tools.write_yaml
merge_yaml = _yaml_tools.merge_yaml
validate_yaml = _yaml_tools.validate_yaml
analyze_yaml = _yaml_tools.analyze_yaml
transform_yaml = _yaml_tools.transform_yaml

if __name__ == "__main__":
    # Example usage
    print("\n==================================================")
    print("YAMLTools Demonstration")
    print("==================================================\n")
    
    # Sample data
    data1 = {
        'server': {
            'host': 'localhost',
            'port': 8080,
            'debug': True
        },
        'database': {
            'url': 'postgresql://localhost:5432/db',
            'pool_size': 5
        }
    }
    
    data2 = {
        'server': {
            'workers': 4,
            'timeout': 30
        },
        'logging': {
            'level': 'INFO',
            'file': 'app.log'
        }
    }
    
    # 1. Write YAML files
    print("1. Writing YAML Files")
    print("------------------------------")
    success1 = write_yaml(data1, 'config1.yaml')
    success2 = write_yaml(data2, 'config2.yaml')
    print(f"First file written: {success1}")
    print(f"Second file written: {success2}")
    print()
    
    # 2. Read YAML file
    print("2. Reading YAML File")
    print("------------------------------")
    config = read_yaml('config1.yaml')
    print("YAML content:")
    print(config)
    print()
    
    # 3. Merge YAML files
    print("3. Merging YAML Files")
    print("------------------------------")
    merged = merge_yaml(['config1.yaml', 'config2.yaml'])
    print("Merged content:")
    print(merged)
    print()
    
    # 4. Analyze YAML
    print("4. Analyzing YAML")
    print("------------------------------")
    analysis = analyze_yaml(merged)
    print("Analysis results:")
    print(analysis)
    print()
    
    # 5. Transform YAML
    print("5. Transforming YAML")
    print("------------------------------")
    operations = [
        {
            'type': 'set',
            'path': 'server/host',
            'value': '0.0.0.0'
        },
        {
            'type': 'delete',
            'path': 'server/debug'
        },
        {
            'type': 'set',
            'path': 'logging/handlers',
            'value': ['console', 'file']
        }
    ]
    transformed = transform_yaml(merged, operations)
    print("Transformed content:")
    print(transformed)
    
    print("\n==================================================")
    print("Demonstration Complete")
    print("==================================================")
    
    # Cleanup
    for file in ['config1.yaml', 'config2.yaml']:
        if os.path.exists(file):
            os.remove(file)



"""
Comprehensive integration test for observability - verifying both explicit and auto-instrumentation patterns.
"""

import pytest
from unittest.mock import MagicMock, patch, call
import os


class TestExplicitTracingPattern:
    """Test the explicit obs.trace()/obs.span() pattern still works."""
    
    def test_explicit_trace_context_manager_works(self):
        """Explicit obs.trace() context manager should work."""
        from praisonai_tools.observability.manager import ObservabilityManager
        
        manager = ObservabilityManager()
        
        # Setup mock provider
        mock_provider = MagicMock()
        mock_provider.init.return_value = True
        mock_provider._initialized = True
        manager._provider = mock_provider
        
        # Test trace context manager (should work even without auto-instrumentation)
        with manager.trace("test-trace") as trace:
            assert trace is not None or True  # trace may be None for mock
            
        # Explicit trace should work
        print("✓ Explicit obs.trace() works")
    
    def test_explicit_span_context_manager_works(self):
        """Explicit obs.span() context manager should work."""
        from praisonai_tools.observability.manager import ObservabilityManager
        from praisonai_tools.observability.base import SpanKind
        
        manager = ObservabilityManager()
        
        # Setup mock provider
        mock_provider = MagicMock()
        mock_provider.init.return_value = True
        mock_provider._initialized = True
        manager._provider = mock_provider
        
        # Test span context manager
        with manager.trace("test-trace"):
            with manager.span("test-span", kind=SpanKind.AGENT) as span:
                assert span is not None or True
        
        print("✓ Explicit obs.span() works")
    
    def test_explicit_pattern_with_auto_instrument_disabled(self):
        """Explicit tracing should work when auto_instrument=False."""
        from praisonai_tools.observability.manager import ObservabilityManager
        
        manager = ObservabilityManager()
        manager._providers['mock'] = MagicMock()
        mock_provider = MagicMock()
        mock_provider.init.return_value = True
        mock_provider._initialized = True
        manager._providers['mock'].return_value = mock_provider
        
        with patch.object(manager, '_load_provider'):
            # Init with auto_instrument=False
            manager.init(provider='mock', auto_instrument=False)
        
        # Explicit trace should still work
        with manager.trace("explicit-trace"):
            pass
        
        print("✓ Explicit tracing works with auto_instrument=False")


class TestAutoInstrumentationPattern:
    """Test the auto-instrumentation pattern."""
    
    def test_auto_instrumentation_patches_agent_class(self):
        """Auto-instrumentation should patch Agent class."""
        from praisonai_tools.observability.manager import ObservabilityManager
        
        manager = ObservabilityManager()
        
        # Setup mock provider
        mock_provider = MagicMock()
        mock_provider.init.return_value = True
        mock_provider._initialized = True
        manager._provider = mock_provider
        
        # Call auto-instrument
        manager._auto_instrument_agents()
        
        # Check if Agent class is patched
        try:
            from praisonaiagents.agent.agent import Agent
            assert getattr(Agent, '_obs_instrumented', False), "Agent class should be instrumented"
            print("✓ Agent class is instrumented")
        except ImportError:
            pytest.skip("praisonaiagents not installed")
    
    def test_auto_instrumentation_default_behavior(self):
        """Auto-instrumentation should be enabled by default on init()."""
        from praisonai_tools.observability.manager import ObservabilityManager
        import inspect
        
        manager = ObservabilityManager()
        
        # Check signature
        sig = inspect.signature(manager.init)
        auto_instrument_param = sig.parameters.get('auto_instrument')
        
        assert auto_instrument_param is not None, "init() should have auto_instrument param"
        assert auto_instrument_param.default == True, "auto_instrument should default to True"
        
        print("✓ auto_instrument defaults to True")
    
    def test_wrap_methods_exist(self):
        """Wrapper methods should exist on ObservabilityManager."""
        from praisonai_tools.observability.manager import ObservabilityManager
        
        manager = ObservabilityManager()
        
        assert hasattr(manager, '_auto_instrument_agents'), "Should have _auto_instrument_agents"
        assert hasattr(manager, '_wrap_agent_class'), "Should have _wrap_agent_class"
        assert hasattr(manager, '_wrap_workflow_class'), "Should have _wrap_workflow_class"
        
        print("✓ All wrapper methods exist")


class TestBothPatternsTogether:
    """Test that both patterns can coexist."""
    
    def test_auto_instrument_then_explicit_trace(self):
        """After auto-instrumentation, explicit trace should still work."""
        from praisonai_tools.observability.manager import ObservabilityManager
        
        manager = ObservabilityManager()
        
        # Setup mock provider and auto-instrument
        mock_provider = MagicMock()
        mock_provider.init.return_value = True
        mock_provider._initialized = True
        manager._provider = mock_provider
        manager._auto_instrument_agents()
        
        # Explicit trace should still work
        with manager.trace("explicit-after-auto"):
            pass
        
        print("✓ Explicit tracing works after auto-instrumentation")


if __name__ == "__main__":
    print("=" * 60)
    print("Observability Integration Test - Both Patterns")
    print("=" * 60)
    
    # Run explicit pattern tests
    print("\n--- Explicit Tracing Pattern ---")
    explicit_tests = TestExplicitTracingPattern()
    explicit_tests.test_explicit_trace_context_manager_works()
    explicit_tests.test_explicit_span_context_manager_works()
    explicit_tests.test_explicit_pattern_with_auto_instrument_disabled()
    
    # Run auto-instrumentation tests
    print("\n--- Auto-Instrumentation Pattern ---")
    auto_tests = TestAutoInstrumentationPattern()
    auto_tests.test_auto_instrumentation_default_behavior()
    auto_tests.test_wrap_methods_exist()
    
    try:
        auto_tests.test_auto_instrumentation_patches_agent_class()
    except Exception as e:
        print(f"⚠ Agent class test skipped: {e}")
    
    # Run coexistence test
    print("\n--- Both Patterns Together ---")
    both_tests = TestBothPatternsTogether()
    both_tests.test_auto_instrument_then_explicit_trace()
    
    print("\n" + "=" * 60)
    print("✅ ALL VERIFICATION TESTS PASSED")
    print("=" * 60)



"""Tests for MediaTool."""

import pytest

from praisonai_tools.recipe_tools.media_tool import MediaTool, media_probe


class TestMediaTool:
    """Unit tests for MediaTool."""
    
    @pytest.fixture
    def tool(self):
        return MediaTool(verbose=True)
    
    @pytest.mark.unit
    def test_check_dependencies(self, tool):
        """Test dependency checking."""
        deps = tool.check_dependencies()
        assert "ffmpeg" in deps
        assert "ffprobe" in deps
    
    @pytest.mark.requires_ffmpeg
    def test_probe_video(self, tool, sample_video, has_ffmpeg):
        """Test probing a video file."""
        if not has_ffmpeg:
            pytest.skip("ffmpeg not available")
        if not sample_video.exists():
            pytest.skip("Sample video not found")
        
        result = tool.probe(sample_video)
        
        assert result.path == str(sample_video)
        assert result.duration > 0
        assert result.has_video or result.has_audio
    
    @pytest.mark.requires_ffmpeg
    def test_probe_audio(self, tool, sample_audio, has_ffmpeg):
        """Test probing an audio file."""
        if not has_ffmpeg:
            pytest.skip("ffmpeg not available")
        if not sample_audio.exists():
            pytest.skip("Sample audio not found")
        
        result = tool.probe(sample_audio)
        
        assert result.path == str(sample_audio)
        assert result.has_audio
    
    @pytest.mark.requires_ffmpeg
    def test_extract_audio(self, tool, sample_video, temp_dir, has_ffmpeg):
        """Test extracting audio from video."""
        if not has_ffmpeg:
            pytest.skip("ffmpeg not available")
        if not sample_video.exists():
            pytest.skip("Sample video not found")
        
        output = temp_dir / "audio.mp3"
        result = tool.extract_audio(sample_video, output)
        
        assert result.exists()
        assert result.stat().st_size > 0
    
    @pytest.mark.requires_ffmpeg
    def test_trim(self, tool, sample_video, temp_dir, has_ffmpeg):
        """Test trimming a video."""
        if not has_ffmpeg:
            pytest.skip("ffmpeg not available")
        if not sample_video.exists():
            pytest.skip("Sample video not found")
        
        output = temp_dir / "trimmed.mp4"
        result = tool.trim(sample_video, output, start=0, duration=2)
        
        assert result.exists()
    
    @pytest.mark.requires_ffmpeg
    def test_extract_frames(self, tool, sample_video, temp_dir, has_ffmpeg):
        """Test extracting frames from video."""
        if not has_ffmpeg:
            pytest.skip("ffmpeg not available")
        if not sample_video.exists():
            pytest.skip("Sample video not found")
        
        frames = tool.extract_frames(sample_video, temp_dir, interval=1)
        
        assert len(frames) > 0
        assert all(f.exists() for f in frames)
    
    @pytest.mark.unit
    def test_probe_file_not_found(self, tool):
        """Test probing non-existent file."""
        with pytest.raises(FileNotFoundError):
            tool.probe("/nonexistent/file.mp4")


class TestMediaToolConvenienceFunctions:
    """Test convenience functions."""
    
    @pytest.mark.requires_ffmpeg
    def test_media_probe(self, sample_video, has_ffmpeg):
        """Test media_probe function."""
        if not has_ffmpeg:
            pytest.skip("ffmpeg not available")
        if not sample_video.exists():
            pytest.skip("Sample video not found")
        
        result = media_probe(sample_video)
        assert result.duration > 0



"""Firestore Tool for PraisonAI Agents.

Google Cloud Firestore operations.

Usage:
    from praisonai_tools import FirestoreTool
    
    fs = FirestoreTool()
    docs = fs.list_documents(collection="users")

Environment Variables:
    GOOGLE_APPLICATION_CREDENTIALS: Path to service account JSON
"""

import os
import logging
from typing import Any, Dict, List, Optional, Union

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)


class FirestoreTool(BaseTool):
    """Tool for Firestore operations."""
    
    name = "firestore"
    description = "Google Cloud Firestore database operations."
    
    def __init__(self, project_id: Optional[str] = None):
        self.project_id = project_id or os.getenv("GOOGLE_CLOUD_PROJECT")
        self._db = None
        super().__init__()
    
    @property
    def db(self):
        if self._db is None:
            try:
                from google.cloud import firestore
            except ImportError:
                raise ImportError("google-cloud-firestore not installed")
            self._db = firestore.Client(project=self.project_id)
        return self._db
    
    def run(
        self,
        action: str = "list",
        collection: Optional[str] = None,
        **kwargs
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        action = action.lower().replace("-", "_")
        
        if action == "list":
            return self.list_documents(collection=collection, **kwargs)
        elif action == "get":
            return self.get_document(collection=collection, doc_id=kwargs.get("doc_id"))
        elif action == "add":
            return self.add_document(collection=collection, data=kwargs.get("data"))
        elif action == "update":
            return self.update_document(collection=collection, doc_id=kwargs.get("doc_id"), data=kwargs.get("data"))
        elif action == "delete":
            return self.delete_document(collection=collection, doc_id=kwargs.get("doc_id"))
        else:
            return {"error": f"Unknown action: {action}"}
    
    def list_documents(self, collection: str, limit: int = 100) -> List[Dict[str, Any]]:
        """List documents in collection."""
        if not collection:
            return [{"error": "collection is required"}]
        
        try:
            docs = self.db.collection(collection).limit(limit).stream()
            return [{"id": doc.id, **doc.to_dict()} for doc in docs]
        except Exception as e:
            logger.error(f"Firestore list error: {e}")
            return [{"error": str(e)}]
    
    def get_document(self, collection: str, doc_id: str) -> Dict[str, Any]:
        """Get document by ID."""
        if not collection or not doc_id:
            return {"error": "collection and doc_id are required"}
        
        try:
            doc = self.db.collection(collection).document(doc_id).get()
            if doc.exists:
                return {"id": doc.id, **doc.to_dict()}
            return {"error": "Document not found"}
        except Exception as e:
            logger.error(f"Firestore get error: {e}")
            return {"error": str(e)}
    
    def add_document(self, collection: str, data: Dict) -> Dict[str, Any]:
        """Add document."""
        if not collection or not data:
            return {"error": "collection and data are required"}
        
        try:
            _, doc_ref = self.db.collection(collection).add(data)
            return {"success": True, "id": doc_ref.id}
        except Exception as e:
            logger.error(f"Firestore add error: {e}")
            return {"error": str(e)}
    
    def update_document(self, collection: str, doc_id: str, data: Dict) -> Dict[str, Any]:
        """Update document."""
        if not collection or not doc_id or not data:
            return {"error": "collection, doc_id, and data are required"}
        
        try:
            self.db.collection(collection).document(doc_id).update(data)
            return {"success": True}
        except Exception as e:
            logger.error(f"Firestore update error: {e}")
            return {"error": str(e)}
    
    def delete_document(self, collection: str, doc_id: str) -> Dict[str, Any]:
        """Delete document."""
        if not collection or not doc_id:
            return {"error": "collection and doc_id are required"}
        
        try:
            self.db.collection(collection).document(doc_id).delete()
            return {"success": True}
        except Exception as e:
            logger.error(f"Firestore delete error: {e}")
            return {"error": str(e)}


def firestore_list(collection: str, limit: int = 100) -> List[Dict[str, Any]]:
    """List Firestore documents."""
    return FirestoreTool().list_documents(collection=collection, limit=limit)



"""
Archive Tool - Zip/tar operations.

Provides:
- Creating archives (zip, tar, tar.gz)
- Extracting archives
- Listing archive contents
- Generating manifests with checksums
"""

import hashlib
import logging
import os
import tarfile
import zipfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Union

import sys
import os as _os
_dir = _os.path.dirname(_os.path.abspath(__file__))
if _dir not in sys.path:
    sys.path.insert(0, _dir)

try:
    from .base import RecipeToolBase
except ImportError:
    from base import RecipeToolBase

logger = logging.getLogger(__name__)


@dataclass
class ArchiveEntry:
    """Entry in an archive."""
    name: str
    size: int
    is_dir: bool
    sha256: Optional[str] = None


@dataclass
class ArchiveManifest:
    """Manifest of an archive."""
    path: str
    format: str
    total_size: int
    file_count: int
    entries: List[ArchiveEntry] = field(default_factory=list)
    
    def to_dict(self) -> Dict:
        """Convert to dictionary."""
        return {
            "path": self.path,
            "format": self.format,
            "total_size": self.total_size,
            "file_count": self.file_count,
            "entries": [
                {"name": e.name, "size": e.size, "is_dir": e.is_dir, "sha256": e.sha256}
                for e in self.entries
            ],
        }


class ArchiveTool(RecipeToolBase):
    """
    Archive operations tool using stdlib.
    
    Provides archive creation, extraction, and listing.
    """
    
    name = "archive_tool"
    description = "Zip/tar archive operations"
    
    # Default exclusions
    DEFAULT_EXCLUDES = [
        ".git",
        ".svn",
        ".hg",
        "__pycache__",
        "*.pyc",
        ".DS_Store",
        "Thumbs.db",
        "node_modules",
        ".venv",
        "venv",
        ".env",
        "*.egg-info",
    ]
    
    def check_dependencies(self) -> Dict[str, bool]:
        """Check for archive support (stdlib, always available)."""
        return {
            "zipfile": True,
            "tarfile": True,
        }
    
    def _should_exclude(self, path: Path, excludes: List[str]) -> bool:
        """Check if path should be excluded."""
        import fnmatch
        
        name = path.name
        for pattern in excludes:
            if fnmatch.fnmatch(name, pattern):
                return True
        return False
    
    def _calculate_sha256(self, path: Path) -> str:
        """Calculate SHA256 hash of a file."""
        sha256 = hashlib.sha256()
        with open(path, "rb") as f:
            for chunk in iter(lambda: f.read(8192), b""):
                sha256.update(chunk)
        return sha256.hexdigest()
    
    def create(
        self,
        input_path: Union[str, Path],
        output_path: Optional[Union[str, Path]] = None,
        format: str = "zip",
        excludes: Optional[List[str]] = None,
        include_checksums: bool = True,
        compression_level: int = 6,
    ) -> Path:
        """
        Create an archive from a file or directory.
        
        Args:
            input_path: Input file or directory
            output_path: Output archive path
            format: Archive format (zip, tar, tar.gz, tar.bz2)
            excludes: Patterns to exclude
            include_checksums: Calculate SHA256 for files
            compression_level: Compression level (0-9)
            
        Returns:
            Path to created archive
        """
        input_path = Path(input_path).resolve()
        
        if not input_path.exists():
            raise FileNotFoundError(f"Input not found: {input_path}")
        
        if excludes is None:
            excludes = self.DEFAULT_EXCLUDES
        
        # Determine output path
        if output_path is None:
            ext = {"zip": ".zip", "tar": ".tar", "tar.gz": ".tar.gz", "tar.bz2": ".tar.bz2"}
            output_path = input_path.parent / f"{input_path.name}{ext.get(format, '.zip')}"
        else:
            output_path = Path(output_path)
            self._ensure_output_dir(output_path.parent)
        
        if format == "zip":
            return self._create_zip(input_path, output_path, excludes, compression_level)
        elif format in ("tar", "tar.gz", "tar.bz2"):
            return self._create_tar(input_path, output_path, format, excludes)
        else:
            raise ValueError(f"Unsupported format: {format}")
    
    def _create_zip(
        self,
        input_path: Path,
        output_path: Path,
        excludes: List[str],
        compression_level: int,
    ) -> Path:
        """Create a zip archive."""
        compression = zipfile.ZIP_DEFLATED
        
        with zipfile.ZipFile(output_path, "w", compression, compresslevel=compression_level) as zf:
            if input_path.is_file():
                zf.write(input_path, input_path.name)
            else:
                for root, dirs, files in os.walk(input_path):
                    # Filter directories in-place
                    dirs[:] = [d for d in dirs if not self._should_exclude(Path(d), excludes)]
                    
                    for file in files:
                        file_path = Path(root) / file
                        if not self._should_exclude(file_path, excludes):
                            arcname = file_path.relative_to(input_path)
                            zf.write(file_path, arcname)
        
        return output_path
    
    def _create_tar(
        self,
        input_path: Path,
        output_path: Path,
        format: str,
        excludes: List[str],
    ) -> Path:
        """Create a tar archive."""
        mode = {"tar": "w", "tar.gz": "w:gz", "tar.bz2": "w:bz2"}[format]
        
        def filter_func(tarinfo):
            if self._should_exclude(Path(tarinfo.name), excludes):
                return None
            return tarinfo
        
        with tarfile.open(output_path, mode) as tf:
            tf.add(input_path, arcname=input_path.name, filter=filter_func)
        
        return output_path
    
    def extract(
        self,
        input_path: Union[str, Path],
        output_dir: Optional[Union[str, Path]] = None,
    ) -> Path:
        """
        Extract an archive.
        
        Args:
            input_path: Archive file
            output_dir: Output directory
            
        Returns:
            Path to extraction directory
        """
        input_path = Path(input_path)
        
        if not input_path.exists():
            raise FileNotFoundError(f"Archive not found: {input_path}")
        
        if output_dir is None:
            output_dir = input_path.parent / input_path.stem
        else:
            output_dir = Path(output_dir)
        
        self._ensure_output_dir(output_dir)
        
        # Detect format
        if input_path.suffix == ".zip":
            with zipfile.ZipFile(input_path, "r") as zf:
                zf.extractall(output_dir)
        elif input_path.suffix in (".tar", ".gz", ".bz2", ".xz"):
            with tarfile.open(input_path, "r:*") as tf:
                tf.extractall(output_dir)
        else:
            raise ValueError(f"Unknown archive format: {input_path.suffix}")
        
        return output_dir
    
    def list(
        self,
        path: Union[str, Path],
        include_checksums: bool = False,
    ) -> ArchiveManifest:
        """
        List contents of an archive.
        
        Args:
            path: Archive file
            include_checksums: Calculate checksums (slower)
            
        Returns:
            ArchiveManifest with entries
        """
        path = Path(path)
        
        if not path.exists():
            raise FileNotFoundError(f"Archive not found: {path}")
        
        entries = []
        total_size = 0
        
        if path.suffix == ".zip":
            format_name = "zip"
            with zipfile.ZipFile(path, "r") as zf:
                for info in zf.infolist():
                    entries.append(ArchiveEntry(
                        name=info.filename,
                        size=info.file_size,
                        is_dir=info.is_dir(),
                    ))
                    total_size += info.file_size
        else:
            format_name = "tar"
            with tarfile.open(path, "r:*") as tf:
                for info in tf.getmembers():
                    entries.append(ArchiveEntry(
                        name=info.name,
                        size=info.size,
                        is_dir=info.isdir(),
                    ))
                    total_size += info.size
        
        return ArchiveManifest(
            path=str(path),
            format=format_name,
            total_size=total_size,
            file_count=len([e for e in entries if not e.is_dir]),
            entries=entries,
        )
    
    def create_manifest(
        self,
        input_path: Union[str, Path],
        output_path: Optional[Union[str, Path]] = None,
        excludes: Optional[List[str]] = None,
    ) -> Dict:
        """
        Create a manifest file for a directory.
        
        Args:
            input_path: Directory to manifest
            output_path: Output JSON file
            excludes: Patterns to exclude
            
        Returns:
            Manifest dictionary
        """
        input_path = Path(input_path).resolve()
        
        if not input_path.is_dir():
            raise NotADirectoryError(f"Not a directory: {input_path}")
        
        if excludes is None:
            excludes = self.DEFAULT_EXCLUDES
        
        entries = []
        total_size = 0
        
        for root, dirs, files in os.walk(input_path):
            # Filter directories
            dirs[:] = [d for d in dirs if not self._should_exclude(Path(d), excludes)]
            
            for file in files:
                file_path = Path(root) / file
                if not self._should_exclude(file_path, excludes):
                    rel_path = file_path.relative_to(input_path)
                    size = file_path.stat().st_size
                    sha256 = self._calculate_sha256(file_path)
                    
                    entries.append({
                        "path": str(rel_path),
                        "size": size,
                        "sha256": sha256,
                    })
                    total_size += size
        
        manifest = {
            "source": str(input_path),
            "created_at": self._get_timestamp(),
            "total_size": total_size,
            "file_count": len(entries),
            "files": entries,
        }
        
        if output_path:
            import json
            output_path = Path(output_path)
            self._ensure_output_dir(output_path.parent)
            output_path.write_text(json.dumps(manifest, indent=2), encoding="utf-8")
        
        return manifest
    
    def create_checksums(
        self,
        input_path: Union[str, Path],
        output_path: Optional[Union[str, Path]] = None,
        algorithm: str = "sha256",
    ) -> str:
        """
        Create checksums file for a directory.
        
        Args:
            input_path: Directory
            output_path: Output checksums file
            algorithm: Hash algorithm (sha256, md5)
            
        Returns:
            Checksums content
        """
        input_path = Path(input_path).resolve()
        
        if not input_path.is_dir():
            raise NotADirectoryError(f"Not a directory: {input_path}")
        
        lines = []
        
        for root, dirs, files in os.walk(input_path):
            dirs[:] = [d for d in dirs if not self._should_exclude(Path(d), self.DEFAULT_EXCLUDES)]
            
            for file in sorted(files):
                file_path = Path(root) / file
                if not self._should_exclude(file_path, self.DEFAULT_EXCLUDES):
                    rel_path = file_path.relative_to(input_path)
                    
                    if algorithm == "sha256":
                        checksum = self._calculate_sha256(file_path)
                    elif algorithm == "md5":
                        md5 = hashlib.md5()
                        with open(file_path, "rb") as f:
                            for chunk in iter(lambda: f.read(8192), b""):
                                md5.update(chunk)
                        checksum = md5.hexdigest()
                    else:
                        raise ValueError(f"Unknown algorithm: {algorithm}")
                    
                    lines.append(f"{checksum}  {rel_path}")
        
        content = "\n".join(lines)
        
        if output_path:
            output_path = Path(output_path)
            self._ensure_output_dir(output_path.parent)
            output_path.write_text(content, encoding="utf-8")
        
        return content


# Convenience functions
def archive_create(
    input_path: Union[str, Path],
    output_path: Optional[Union[str, Path]] = None,
    format: str = "zip",
    verbose: bool = False,
) -> Path:
    """Create an archive."""
    return ArchiveTool(verbose=verbose).create(input_path, output_path, format)


def archive_extract(
    input_path: Union[str, Path],
    output_dir: Optional[Union[str, Path]] = None,
    verbose: bool = False,
) -> Path:
    """Extract an archive."""
    return ArchiveTool(verbose=verbose).extract(input_path, output_dir)


def archive_list(
    path: Union[str, Path],
    verbose: bool = False,
) -> ArchiveManifest:
    """List archive contents."""
    return ArchiveTool(verbose=verbose).list(path)



"""Mem0 Tool for PraisonAI Agents.

Memory management using Mem0.

Usage:
    from praisonai_tools import Mem0Tool
    
    mem0 = Mem0Tool()
    mem0.add("User prefers dark mode")

Environment Variables:
    MEM0_API_KEY: Mem0 API key (optional for local)
"""

import os
import logging
from typing import Any, Dict, List, Optional, Union

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)


class Mem0Tool(BaseTool):
    """Tool for Mem0 memory management."""
    
    name = "mem0"
    description = "Store and retrieve memories using Mem0."
    
    def __init__(self, api_key: Optional[str] = None):
        self.api_key = api_key or os.getenv("MEM0_API_KEY")
        self._memory = None
        super().__init__()
    
    @property
    def memory(self):
        if self._memory is None:
            try:
                from mem0 import Memory
            except ImportError:
                raise ImportError("mem0ai not installed. Install with: pip install mem0ai")
            self._memory = Memory()
        return self._memory
    
    def run(
        self,
        action: str = "add",
        text: Optional[str] = None,
        user_id: str = "default",
        **kwargs
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        action = action.lower().replace("-", "_")
        
        if action == "add":
            return self.add(text=text, user_id=user_id)
        elif action == "search":
            return self.search(query=text, user_id=user_id)
        elif action == "get_all":
            return self.get_all(user_id=user_id)
        else:
            return {"error": f"Unknown action: {action}"}
    
    def add(self, text: str, user_id: str = "default") -> Dict[str, Any]:
        """Add memory."""
        if not text:
            return {"error": "text is required"}
        
        try:
            result = self.memory.add(text, user_id=user_id)
            return {"success": True, "result": result}
        except Exception as e:
            logger.error(f"Mem0 add error: {e}")
            return {"error": str(e)}
    
    def search(self, query: str, user_id: str = "default") -> List[Dict[str, Any]]:
        """Search memories."""
        if not query:
            return [{"error": "query is required"}]
        
        try:
            results = self.memory.search(query, user_id=user_id)
            return results
        except Exception as e:
            logger.error(f"Mem0 search error: {e}")
            return [{"error": str(e)}]
    
    def get_all(self, user_id: str = "default") -> List[Dict[str, Any]]:
        """Get all memories."""
        try:
            results = self.memory.get_all(user_id=user_id)
            return results
        except Exception as e:
            logger.error(f"Mem0 get_all error: {e}")
            return [{"error": str(e)}]


def mem0_add(text: str, user_id: str = "default") -> Dict[str, Any]:
    """Add memory to Mem0."""
    return Mem0Tool().add(text=text, user_id=user_id)



"""CSV Tool for PraisonAI Agents.

CSV file operations.

Usage:
    from praisonai_tools import CSVTool
    
    csv = CSVTool()
    data = csv.read("data.csv")
"""

import logging
from typing import Any, Dict, List, Optional, Union

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)


class CSVTool(BaseTool):
    """Tool for CSV file operations."""
    
    name = "csv"
    description = "Read and write CSV files."
    
    def run(
        self,
        action: str = "read",
        file_path: Optional[str] = None,
        **kwargs
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        action = action.lower().replace("-", "_")
        
        if action == "read":
            return self.read(file_path=file_path, **kwargs)
        elif action == "write":
            return self.write(file_path=file_path, data=kwargs.get("data"))
        elif action == "query":
            return self.query(file_path=file_path, sql=kwargs.get("sql"))
        else:
            return {"error": f"Unknown action: {action}"}
    
    def read(self, file_path: str, limit: int = 100) -> List[Dict[str, Any]]:
        """Read CSV file."""
        if not file_path:
            return [{"error": "file_path is required"}]
        
        try:
            import csv
            with open(file_path, "r", newline="", encoding="utf-8") as f:
                reader = csv.DictReader(f)
                rows = []
                for i, row in enumerate(reader):
                    if i >= limit:
                        break
                    rows.append(dict(row))
                return rows
        except Exception as e:
            logger.error(f"CSV read error: {e}")
            return [{"error": str(e)}]
    
    def write(self, file_path: str, data: List[Dict]) -> Dict[str, Any]:
        """Write CSV file."""
        if not file_path or not data:
            return {"error": "file_path and data are required"}
        
        try:
            import csv
            fieldnames = list(data[0].keys())
            with open(file_path, "w", newline="", encoding="utf-8") as f:
                writer = csv.DictWriter(f, fieldnames=fieldnames)
                writer.writeheader()
                writer.writerows(data)
            return {"success": True, "rows_written": len(data)}
        except Exception as e:
            logger.error(f"CSV write error: {e}")
            return {"error": str(e)}
    
    def query(self, file_path: str, sql: str) -> List[Dict[str, Any]]:
        """Query CSV with SQL using DuckDB."""
        if not file_path or not sql:
            return [{"error": "file_path and sql are required"}]
        
        try:
            import duckdb
            result = duckdb.query(sql.replace("$FILE", f"'{file_path}'")).fetchdf()
            return result.to_dict(orient="records")
        except ImportError:
            return [{"error": "duckdb not installed for SQL queries"}]
        except Exception as e:
            logger.error(f"CSV query error: {e}")
            return [{"error": str(e)}]


def read_csv(file_path: str, limit: int = 100) -> List[Dict[str, Any]]:
    """Read CSV file."""
    return CSVTool().read(file_path=file_path, limit=limit)



"""
LLM Prompting for Final Cut Pro Integration

Converts natural language instructions to EditIntent JSON using OpenAI SDK.
"""

from __future__ import annotations

import json
import os
from typing import Any, Optional

from pydantic import ValidationError

from .intent import (
    Asset,
    AudioRole,
    AudioSettings,
    EditIntent,
    Project,
    ProjectFormat,
    Segment,
    Timeline,
    get_format_preset,
)


EDIT_INTENT_SCHEMA = EditIntent.model_json_schema()

SYSTEM_PROMPT = """You are an expert video editor assistant that converts natural language editing instructions into structured JSON for Final Cut Pro.

You MUST output ONLY valid JSON that conforms to the EditIntent schema. No prose, no explanations, just JSON.

## CRITICAL RULES:
1. ONLY use media files from the provided media_paths list. NEVER invent file paths.
2. If the user requests media not in the list, add it to "missing_inputs" array.
3. For time-based edits without specific timestamps, set "needs_user_timestamps": true and use default concatenation order.
4. Use rational time format for all time values: "numerator/denominators" (e.g., "2500/2500s" = 1 second at 25fps)
5. Default timescale is fps * 100 (e.g., 25fps → 2500)

## CAPABILITIES (v1):
- Create projects with primary storyline
- Place asset clips in sequence
- Set audio roles (dialogue, music, effects)
- Add chapter markers
- Basic volume adjustments

## NOT IMPLEMENTED (store in operations but will warn):
- remove_pauses_over_seconds: Silence/pause removal
- loudness_target_lufs: Audio normalization
- zoom_punches: Ken Burns / zoom effects

## SCHEMA STRUCTURE:
{schema}

## EXAMPLE OUTPUT:
```json
{{
  "version": "1",
  "project": {{
    "name": "My Edit",
    "format": {{"width": 1920, "height": 1080, "fps": 25.0}},
    "audio": {{"layout": "stereo", "rate": 48000, "channels": 2}}
  }},
  "assets": [
    {{
      "id": "r2",
      "name": "interview",
      "path": "/path/to/interview.mov",
      "has_video": true,
      "has_audio": true
    }}
  ],
  "timeline": {{
    "segments": [
      {{
        "asset_id": "r2",
        "offset": "0/2500s",
        "start": "0/2500s",
        "duration": "5000/2500s",
        "lane": 0,
        "role": "dialogue"
      }}
    ],
    "markers": []
  }}
}}
```
"""

USER_PROMPT_TEMPLATE = """## INSTRUCTION:
{instruction}

## AVAILABLE MEDIA FILES:
{media_list}

## PROJECT SETTINGS:
- Project name: {project_name}
- Format: {format_info}

Generate the EditIntent JSON now. Output ONLY the JSON, no other text."""


def _get_openai_client():
    """Lazy import and create OpenAI client."""
    try:
        from openai import OpenAI
    except ImportError:
        raise ImportError(
            "OpenAI SDK not installed. Install with: pip install openai"
        )

    api_key = os.environ.get("OPENAI_API_KEY")
    if not api_key:
        raise ValueError(
            "OPENAI_API_KEY environment variable not set. "
            "Set it with: export OPENAI_API_KEY=your-key"
        )

    return OpenAI(api_key=api_key)


def generate_edit_intent(
    instruction: str,
    media_paths: list[str],
    project_name: str = "Untitled Project",
    format_preset: Optional[str] = None,
    width: Optional[int] = None,
    height: Optional[int] = None,
    fps: Optional[float] = None,
    model: str = "gpt-4o",
    max_retries: int = 3,
    agent_id: Optional[str] = None,
    run_id: Optional[str] = None,
    session_id: Optional[str] = None,
) -> tuple[EditIntent, list[str]]:
    """
    Generate EditIntent from natural language instruction using OpenAI.
    
    Args:
        instruction: Natural language editing instruction
        media_paths: List of available media file paths
        project_name: Name for the project
        format_preset: Format preset name (e.g., "1080p25")
        width: Custom width (overrides preset)
        height: Custom height (overrides preset)
        fps: Custom fps (overrides preset)
        model: OpenAI model to use
        max_retries: Maximum retry attempts for validation
        agent_id: Agent ID for attribution
        run_id: Run ID for attribution
        session_id: Session ID for attribution
        
    Returns:
        Tuple of (EditIntent, warnings)
    """
    client = _get_openai_client()
    warnings = []

    for path in media_paths:
        if not os.path.isabs(path):
            raise ValueError(f"Media path must be absolute: {path}")
        if not os.path.exists(path):
            warnings.append(f"Media file not found (will be referenced anyway): {path}")

    if format_preset:
        project_format = get_format_preset(format_preset)
    else:
        project_format = ProjectFormat(
            width=width or 1920,
            height=height or 1080,
            fps=fps or 25.0,
        )

    if width:
        project_format.width = width
    if height:
        project_format.height = height
    if fps:
        project_format.fps = fps

    media_list = "\n".join(f"- {path}" for path in media_paths) or "No media files provided"
    format_info = f"{project_format.width}x{project_format.height} @ {project_format.fps}fps"

    system_prompt = SYSTEM_PROMPT.format(schema=json.dumps(EDIT_INTENT_SCHEMA, indent=2))
    user_prompt = USER_PROMPT_TEMPLATE.format(
        instruction=instruction,
        media_list=media_list,
        project_name=project_name,
        format_info=format_info,
    )

    last_error = None
    for attempt in range(max_retries):
        try:
            if attempt > 0:
                user_prompt = f"{user_prompt}\n\n## PREVIOUS ERROR (attempt {attempt}):\n{last_error}\n\nPlease fix the JSON and try again."

            response = client.chat.completions.create(
                model=model,
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": user_prompt},
                ],
                response_format={"type": "json_object"},
                temperature=0.1,
            )

            content = response.choices[0].message.content
            if not content:
                last_error = "Empty response from LLM"
                continue

            try:
                data = json.loads(content)
            except json.JSONDecodeError as e:
                last_error = f"Invalid JSON: {e}"
                continue

            data = _normalize_intent_data(data, media_paths, project_name, project_format)

            intent = EditIntent.model_validate(data)

            intent.agent_id = agent_id
            intent.run_id = run_id
            intent.session_id = session_id

            warnings.extend(intent.get_warnings())

            return intent, warnings

        except ValidationError as e:
            last_error = f"Schema validation error: {e}"
            continue
        except Exception as e:
            last_error = f"Error: {e}"
            continue

    raise ValueError(
        f"Failed to generate valid EditIntent after {max_retries} attempts. "
        f"Last error: {last_error}"
    )


def _normalize_intent_data(
    data: dict[str, Any],
    media_paths: list[str],
    project_name: str,
    project_format: ProjectFormat,
) -> dict[str, Any]:
    """
    Normalize and fill in missing fields in the intent data.
    
    Args:
        data: Raw data from LLM
        media_paths: Available media paths
        project_name: Project name
        project_format: Project format settings
        
    Returns:
        Normalized data dictionary
    """
    if "version" not in data:
        data["version"] = "1"

    if "project" not in data:
        data["project"] = {}

    if "name" not in data["project"]:
        data["project"]["name"] = project_name

    if "format" not in data["project"]:
        data["project"]["format"] = project_format.model_dump()
    else:
        fmt = data["project"]["format"]
        if "width" not in fmt:
            fmt["width"] = project_format.width
        if "height" not in fmt:
            fmt["height"] = project_format.height
        if "fps" not in fmt:
            fmt["fps"] = project_format.fps

    if "audio" not in data["project"]:
        data["project"]["audio"] = {"layout": "stereo", "rate": 48000, "channels": 2}

    if "assets" not in data or not data["assets"]:
        data["assets"] = []
        for i, path in enumerate(media_paths):
            ext = os.path.splitext(path)[1].lower()
            has_video = ext in (".mov", ".mp4", ".m4v", ".avi", ".mkv", ".webm")
            has_audio = ext in (".wav", ".mp3", ".aac", ".m4a", ".aiff") or has_video

            data["assets"].append({
                "id": f"r{i + 2}",
                "name": os.path.basename(path),
                "path": path,
                "has_video": has_video,
                "has_audio": has_audio,
            })

    for asset in data["assets"]:
        if "id" not in asset:
            asset["id"] = f"r{data['assets'].index(asset) + 2}"
        if "name" not in asset:
            asset["name"] = os.path.basename(asset.get("path", "unknown"))

    if "timeline" not in data:
        data["timeline"] = {"segments": [], "markers": []}

    if "segments" not in data["timeline"]:
        data["timeline"]["segments"] = []

    if "markers" not in data["timeline"]:
        data["timeline"]["markers"] = []

    if not data["timeline"]["segments"] and data["assets"]:
        fps = data["project"]["format"].get("fps", 25.0)
        timescale = int(fps * 100)
        offset = 0

        for asset in data["assets"]:
            duration_seconds = 10.0
            duration_frames = int(duration_seconds * timescale)

            data["timeline"]["segments"].append({
                "asset_id": asset["id"],
                "offset": f"{offset}/{timescale}s",
                "start": f"0/{timescale}s",
                "duration": f"{duration_frames}/{timescale}s",
                "lane": 0,
            })
            offset += duration_frames

    return data


def create_simple_intent(
    media_paths: list[str],
    project_name: str = "Simple Project",
    format_preset: str = "1080p25",
) -> EditIntent:
    """
    Create a simple EditIntent that concatenates all media files.
    
    This is useful for testing or when LLM is not available.
    
    Args:
        media_paths: List of media file paths
        project_name: Project name
        format_preset: Format preset name
        
    Returns:
        EditIntent with all media concatenated
    """
    project_format = get_format_preset(format_preset)
    timescale = project_format.get_timescale()

    assets = []
    segments = []
    offset = 0

    for i, path in enumerate(media_paths):
        if not os.path.isabs(path):
            raise ValueError(f"Media path must be absolute: {path}")

        ext = os.path.splitext(path)[1].lower()
        has_video = ext in (".mov", ".mp4", ".m4v", ".avi", ".mkv", ".webm")
        has_audio = ext in (".wav", ".mp3", ".aac", ".m4a", ".aiff") or has_video

        asset_id = f"r{i + 2}"
        assets.append(Asset(
            id=asset_id,
            name=os.path.basename(path),
            path=path,
            has_video=has_video,
            has_audio=has_audio,
            format_ref="r1" if has_video else None,
        ))

        duration_frames = 10 * timescale

        segments.append(Segment(
            asset_id=asset_id,
            offset=f"{offset}/{timescale}s",
            start=f"0/{timescale}s",
            duration=f"{duration_frames}/{timescale}s",
            lane=0,
            role=AudioRole.DIALOGUE if has_audio else None,
        ))
        offset += duration_frames

    return EditIntent(
        version="1",
        project=Project(
            name=project_name,
            format=project_format,
            audio=AudioSettings(),
        ),
        assets=assets,
        timeline=Timeline(segments=segments, markers=[]),
    )



"""
PraisonAI Documentation Generator.

This module can be run standalone without requiring the praisonaiagents dependency.

Usage:
    python -m praisonai_tools.docs_generator --package rust
    python -m praisonai_tools.docs_generator --package all
"""

# This file is intentionally minimal to allow standalone execution
# without importing the parent package dependencies.

__version__ = "0.1.0"



"""Cassandra Tool for PraisonAI Agents.

Apache Cassandra database operations.

Usage:
    from praisonai_tools import CassandraTool
    
    cassandra = CassandraTool()
    results = cassandra.query("SELECT * FROM users")

Environment Variables:
    CASSANDRA_HOST: Cassandra host
    CASSANDRA_PORT: Cassandra port
    CASSANDRA_KEYSPACE: Keyspace name
"""

import os
import logging
from typing import Any, Dict, List, Optional, Union

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)


class CassandraTool(BaseTool):
    """Tool for Cassandra operations."""
    
    name = "cassandra"
    description = "Apache Cassandra database operations."
    
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        keyspace: Optional[str] = None,
    ):
        self.host = host or os.getenv("CASSANDRA_HOST", "localhost")
        self.port = port or int(os.getenv("CASSANDRA_PORT", "9042"))
        self.keyspace = keyspace or os.getenv("CASSANDRA_KEYSPACE")
        self._session = None
        super().__init__()
    
    @property
    def session(self):
        if self._session is None:
            try:
                from cassandra.cluster import Cluster
            except ImportError:
                raise ImportError("cassandra-driver not installed")
            cluster = Cluster([self.host], port=self.port)
            self._session = cluster.connect(self.keyspace)
        return self._session
    
    def run(
        self,
        action: str = "query",
        cql: Optional[str] = None,
        **kwargs
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        action = action.lower().replace("-", "_")
        
        if action == "query":
            return self.query(cql=cql)
        elif action == "execute":
            return self.execute(cql=cql)
        else:
            return {"error": f"Unknown action: {action}"}
    
    def query(self, cql: str) -> List[Dict[str, Any]]:
        """Execute CQL query."""
        if not cql:
            return [{"error": "cql is required"}]
        
        try:
            rows = self.session.execute(cql)
            return [dict(row._asdict()) for row in rows]
        except Exception as e:
            logger.error(f"Cassandra query error: {e}")
            return [{"error": str(e)}]
    
    def execute(self, cql: str) -> Dict[str, Any]:
        """Execute CQL statement."""
        if not cql:
            return {"error": "cql is required"}
        
        try:
            self.session.execute(cql)
            return {"success": True}
        except Exception as e:
            logger.error(f"Cassandra execute error: {e}")
            return {"error": str(e)}


def cassandra_query(cql: str) -> List[Dict[str, Any]]:
    """Query Cassandra."""
    return CassandraTool().query(cql=cql)



"""Base classes for PraisonAI Tools.

This module re-exports BaseTool and related classes from praisonaiagents.
This ensures a single source of truth for the base tool infrastructure.

Usage:
    from praisonai_tools import BaseTool

    class MyTool(BaseTool):
        name = "my_tool"
        description = "Does something useful"
        
        def run(self, query: str) -> str:
            return f"Result for {query}"
"""

# Import from praisonaiagents - single source of truth
from praisonaiagents.tools.base import (
    BaseTool,
    ToolResult,
    ToolValidationError,
    validate_tool,
)

__all__ = [
    "BaseTool",
    "ToolResult",
    "ToolValidationError",
    "validate_tool",
]



"""Twilio Tool for PraisonAI Agents.

Send SMS and make calls using Twilio.

Usage:
    from praisonai_tools import TwilioTool
    
    twilio = TwilioTool()
    twilio.send_sms(to="+1234567890", body="Hello!")

Environment Variables:
    TWILIO_ACCOUNT_SID: Twilio Account SID
    TWILIO_AUTH_TOKEN: Twilio Auth Token
    TWILIO_PHONE_NUMBER: Default from phone number
"""

import os
import logging
from typing import Any, Dict, Optional, Union

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)


class TwilioTool(BaseTool):
    """Tool for Twilio SMS and calls."""
    
    name = "twilio"
    description = "Send SMS messages and make calls using Twilio."
    
    def __init__(
        self,
        account_sid: Optional[str] = None,
        auth_token: Optional[str] = None,
        from_number: Optional[str] = None,
    ):
        self.account_sid = account_sid or os.getenv("TWILIO_ACCOUNT_SID")
        self.auth_token = auth_token or os.getenv("TWILIO_AUTH_TOKEN")
        self.from_number = from_number or os.getenv("TWILIO_PHONE_NUMBER")
        self._client = None
        super().__init__()
    
    @property
    def client(self):
        if self._client is None:
            try:
                from twilio.rest import Client
            except ImportError:
                raise ImportError("twilio not installed. Install with: pip install twilio")
            
            if not self.account_sid or not self.auth_token:
                raise ValueError("TWILIO_ACCOUNT_SID and TWILIO_AUTH_TOKEN required")
            
            self._client = Client(self.account_sid, self.auth_token)
        return self._client
    
    def run(
        self,
        action: str = "send_sms",
        to: Optional[str] = None,
        body: Optional[str] = None,
        **kwargs
    ) -> Union[str, Dict[str, Any]]:
        action = action.lower().replace("-", "_")
        
        if action == "send_sms":
            return self.send_sms(to=to, body=body)
        elif action == "send_whatsapp":
            return self.send_whatsapp(to=to, body=body)
        elif action == "make_call":
            return self.make_call(to=to, **kwargs)
        elif action == "get_messages":
            return self.get_messages(**kwargs)
        else:
            return {"error": f"Unknown action: {action}"}
    
    def send_sms(self, to: str, body: str, from_number: Optional[str] = None) -> Dict[str, Any]:
        """Send SMS message."""
        if not to or not body:
            return {"error": "to and body are required"}
        
        try:
            message = self.client.messages.create(
                body=body,
                from_=from_number or self.from_number,
                to=to,
            )
            return {
                "success": True,
                "sid": message.sid,
                "status": message.status,
                "to": to,
            }
        except Exception as e:
            logger.error(f"Twilio send_sms error: {e}")
            return {"error": str(e)}
    
    def send_whatsapp(self, to: str, body: str) -> Dict[str, Any]:
        """Send WhatsApp message."""
        if not to or not body:
            return {"error": "to and body are required"}
        
        try:
            message = self.client.messages.create(
                body=body,
                from_=f"whatsapp:{self.from_number}",
                to=f"whatsapp:{to}",
            )
            return {
                "success": True,
                "sid": message.sid,
                "status": message.status,
            }
        except Exception as e:
            logger.error(f"Twilio send_whatsapp error: {e}")
            return {"error": str(e)}
    
    def make_call(self, to: str, twiml: Optional[str] = None, url: Optional[str] = None) -> Dict[str, Any]:
        """Make a phone call."""
        if not to:
            return {"error": "to is required"}
        if not twiml and not url:
            return {"error": "twiml or url is required"}
        
        try:
            params = {
                "to": to,
                "from_": self.from_number,
            }
            if twiml:
                params["twiml"] = twiml
            if url:
                params["url"] = url
            
            call = self.client.calls.create(**params)
            return {
                "success": True,
                "sid": call.sid,
                "status": call.status,
            }
        except Exception as e:
            logger.error(f"Twilio make_call error: {e}")
            return {"error": str(e)}
    
    def get_messages(self, limit: int = 20) -> list:
        """Get recent messages."""
        try:
            messages = self.client.messages.list(limit=limit)
            return [
                {
                    "sid": m.sid,
                    "from": m.from_,
                    "to": m.to,
                    "body": m.body,
                    "status": m.status,
                    "date_sent": str(m.date_sent),
                }
                for m in messages
            ]
        except Exception as e:
            logger.error(f"Twilio get_messages error: {e}")
            return [{"error": str(e)}]


def send_sms(to: str, body: str) -> Dict[str, Any]:
    """Send SMS via Twilio."""
    return TwilioTool().send_sms(to=to, body=body)



"""Pinecone Vector DB Tool for PraisonAI Agents.

Vector database operations using Pinecone.

Usage:
    from praisonai_tools import PineconeTool
    
    pc = PineconeTool()
    pc.upsert(vectors=[{"id": "1", "values": [...], "metadata": {...}}])

Environment Variables:
    PINECONE_API_KEY: Pinecone API key
    PINECONE_INDEX: Default index name
"""

import os
import logging
from typing import Any, Dict, List, Optional, Union

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)


class PineconeTool(BaseTool):
    """Tool for Pinecone vector database."""
    
    name = "pinecone"
    description = "Store and query vectors in Pinecone."
    
    def __init__(
        self,
        api_key: Optional[str] = None,
        index_name: Optional[str] = None,
    ):
        self.api_key = api_key or os.getenv("PINECONE_API_KEY")
        self.index_name = index_name or os.getenv("PINECONE_INDEX")
        self._client = None
        self._index = None
        super().__init__()
    
    @property
    def client(self):
        if self._client is None:
            try:
                from pinecone import Pinecone
            except ImportError:
                raise ImportError("pinecone-client not installed. Install with: pip install pinecone-client")
            
            if not self.api_key:
                raise ValueError("PINECONE_API_KEY required")
            
            self._client = Pinecone(api_key=self.api_key)
        return self._client
    
    @property
    def index(self):
        if self._index is None:
            if not self.index_name:
                raise ValueError("index_name required")
            self._index = self.client.Index(self.index_name)
        return self._index
    
    def run(
        self,
        action: str = "query",
        vector: Optional[List[float]] = None,
        vectors: Optional[List[Dict]] = None,
        ids: Optional[List[str]] = None,
        top_k: int = 10,
        **kwargs
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        action = action.lower().replace("-", "_")
        
        if action == "query":
            return self.query(vector=vector, top_k=top_k, **kwargs)
        elif action == "upsert":
            return self.upsert(vectors=vectors)
        elif action == "delete":
            return self.delete(ids=ids)
        elif action == "describe_index":
            return self.describe_index()
        elif action == "list_indexes":
            return self.list_indexes()
        else:
            return {"error": f"Unknown action: {action}"}
    
    def query(
        self,
        vector: List[float],
        top_k: int = 10,
        include_metadata: bool = True,
        filter_dict: Optional[Dict] = None,
    ) -> List[Dict[str, Any]]:
        """Query similar vectors."""
        if not vector:
            return [{"error": "vector is required"}]
        
        try:
            params = {
                "vector": vector,
                "top_k": top_k,
                "include_metadata": include_metadata,
            }
            if filter_dict:
                params["filter"] = filter_dict
            
            result = self.index.query(**params)
            
            matches = []
            for match in result.get("matches", []):
                matches.append({
                    "id": match["id"],
                    "score": match["score"],
                    "metadata": match.get("metadata", {}),
                })
            return matches
        except Exception as e:
            logger.error(f"Pinecone query error: {e}")
            return [{"error": str(e)}]
    
    def upsert(self, vectors: List[Dict]) -> Dict[str, Any]:
        """Upsert vectors."""
        if not vectors:
            return {"error": "vectors is required"}
        
        try:
            result = self.index.upsert(vectors=vectors)
            return {"success": True, "upserted_count": result.get("upserted_count")}
        except Exception as e:
            logger.error(f"Pinecone upsert error: {e}")
            return {"error": str(e)}
    
    def delete(self, ids: List[str]) -> Dict[str, Any]:
        """Delete vectors by ID."""
        if not ids:
            return {"error": "ids is required"}
        
        try:
            self.index.delete(ids=ids)
            return {"success": True, "deleted": len(ids)}
        except Exception as e:
            logger.error(f"Pinecone delete error: {e}")
            return {"error": str(e)}
    
    def describe_index(self) -> Dict[str, Any]:
        """Get index stats."""
        try:
            stats = self.index.describe_index_stats()
            return {
                "dimension": stats.get("dimension"),
                "total_vector_count": stats.get("total_vector_count"),
                "namespaces": stats.get("namespaces", {}),
            }
        except Exception as e:
            logger.error(f"Pinecone describe_index error: {e}")
            return {"error": str(e)}
    
    def list_indexes(self) -> List[Dict[str, Any]]:
        """List all indexes."""
        try:
            indexes = self.client.list_indexes()
            return [{"name": idx.name, "dimension": idx.dimension} for idx in indexes]
        except Exception as e:
            logger.error(f"Pinecone list_indexes error: {e}")
            return [{"error": str(e)}]


def pinecone_query(vector: List[float], top_k: int = 10) -> List[Dict[str, Any]]:
    """Query Pinecone."""
    return PineconeTool().query(vector=vector, top_k=top_k)



"""AWS SES Tool for PraisonAI Agents.

Send emails via AWS SES.

Usage:
    from praisonai_tools import AWSSESTool
    
    ses = AWSSESTool()
    ses.send_email(to="user@example.com", subject="Hello", body="Hi!")

Environment Variables:
    AWS_ACCESS_KEY_ID: AWS access key
    AWS_SECRET_ACCESS_KEY: AWS secret key
    AWS_REGION: AWS region
"""

import os
import logging
from typing import Any, Dict, Optional, Union

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)


class AWSSESTool(BaseTool):
    """Tool for AWS SES email."""
    
    name = "aws_ses"
    description = "Send emails via AWS SES."
    
    def __init__(self, region: Optional[str] = None):
        self.region = region or os.getenv("AWS_REGION", "us-east-1")
        self._client = None
        super().__init__()
    
    @property
    def client(self):
        if self._client is None:
            try:
                import boto3
            except ImportError:
                raise ImportError("boto3 not installed")
            self._client = boto3.client("ses", region_name=self.region)
        return self._client
    
    def run(
        self,
        action: str = "send",
        to: Optional[str] = None,
        subject: Optional[str] = None,
        body: Optional[str] = None,
        **kwargs
    ) -> Union[str, Dict[str, Any]]:
        if action == "send":
            return self.send_email(to=to, subject=subject, body=body, **kwargs)
        return {"error": f"Unknown action: {action}"}
    
    def send_email(self, to: str, subject: str, body: str, from_email: str = None) -> Dict[str, Any]:
        """Send email."""
        if not to or not subject or not body:
            return {"error": "to, subject, and body are required"}
        
        try:
            resp = self.client.send_email(
                Source=from_email or os.getenv("AWS_SES_FROM_EMAIL", to),
                Destination={"ToAddresses": [to]},
                Message={
                    "Subject": {"Data": subject},
                    "Body": {"Text": {"Data": body}},
                },
            )
            return {"success": True, "message_id": resp.get("MessageId")}
        except Exception as e:
            logger.error(f"SES send_email error: {e}")
            return {"error": str(e)}


def ses_send_email(to: str, subject: str, body: str) -> Dict[str, Any]:
    """Send email via SES."""
    return AWSSESTool().send_email(to=to, subject=subject, body=body)



"""Line Tool for PraisonAI Agents.

Send messages via LINE Messaging API.

Usage:
    from praisonai_tools import LineTool
    
    line = LineTool()
    line.send_message(to="user_id", message="Hello from AI!")

Environment Variables:
    LINE_CHANNEL_ACCESS_TOKEN: LINE Channel access token
    LINE_CHANNEL_SECRET: LINE Channel secret (for webhook verification)

Setup:
    1. Create a LINE Messaging API channel at https://developers.line.biz/
    2. Get your Channel access token and Channel secret
    3. Set environment variables
"""

import os
import logging
import hashlib
import hmac
import base64
from typing import Any, Dict, List, Optional, Union
from dataclasses import dataclass

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)

# LINE API endpoints
LINE_API_BASE = "https://api.line.me/v2"
LINE_API_DATA = "https://api-data.line.me/v2"


@dataclass
class LineMessageResult:
    """Result of sending a LINE message."""
    success: bool
    message_id: Optional[str] = None
    error: Optional[str] = None


class LineTool(BaseTool):
    """Tool for LINE messaging via Messaging API.
    
    Supports:
    - Push messages to users
    - Reply messages (with reply token)
    - Multicast to multiple users
    - Broadcast to all followers
    - Rich messages (Flex Messages)
    - Quick replies
    
    Based on moltbot's LINE implementation patterns.
    """
    
    name = "line"
    description = "Send messages via LINE Messaging API."
    
    def __init__(
        self,
        channel_access_token: Optional[str] = None,
        channel_secret: Optional[str] = None,
    ):
        """Initialize LineTool.
        
        Args:
            channel_access_token: LINE Channel access token (or use LINE_CHANNEL_ACCESS_TOKEN env var)
            channel_secret: LINE Channel secret (or use LINE_CHANNEL_SECRET env var)
        """
        self.channel_access_token = (
            channel_access_token 
            or os.getenv("LINE_CHANNEL_ACCESS_TOKEN") 
            or os.getenv("LINE_ACCESS_TOKEN")
        )
        self.channel_secret = (
            channel_secret 
            or os.getenv("LINE_CHANNEL_SECRET")
        )
        super().__init__()
    
    def _request(
        self,
        method: str,
        endpoint: str,
        data: Optional[Dict] = None,
        base_url: str = LINE_API_BASE,
    ) -> Dict[str, Any]:
        """Make LINE API request."""
        try:
            import requests
        except ImportError:
            return {"error": "requests not installed. Install with: pip install requests"}
        
        if not self.channel_access_token:
            return {"error": "LINE_CHANNEL_ACCESS_TOKEN not configured"}
        
        url = f"{base_url}{endpoint}"
        headers = {
            "Authorization": f"Bearer {self.channel_access_token}",
            "Content-Type": "application/json",
        }
        
        try:
            if method.upper() == "GET":
                response = requests.get(url, headers=headers, timeout=10)
            elif method.upper() == "DELETE":
                response = requests.delete(url, headers=headers, timeout=10)
            else:
                response = requests.post(url, headers=headers, json=data, timeout=10)
            
            if response.status_code == 200:
                if response.text:
                    return response.json()
                return {"success": True}
            
            if response.text:
                error_data = response.json()
                return {"error": error_data.get("message", str(error_data))}
            
            return {"error": f"HTTP {response.status_code}"}
        except Exception as e:
            logger.error(f"LINE API error: {e}")
            return {"error": str(e)}
    
    def run(
        self,
        action: str = "push",
        to: Optional[str] = None,
        message: Optional[str] = None,
        **kwargs
    ) -> Union[str, Dict[str, Any]]:
        """Execute LINE action.
        
        Args:
            action: "push", "reply", "multicast", "broadcast", "get_profile"
            to: User ID, group ID, or room ID
            message: Message text
        """
        action = action.lower().replace("-", "_")
        
        if action == "push":
            return self.push_message(to=to, message=message, **kwargs)
        elif action == "reply":
            return self.reply_message(reply_token=kwargs.get("reply_token"), message=message, **kwargs)
        elif action == "multicast":
            return self.multicast(user_ids=kwargs.get("user_ids", []), message=message, **kwargs)
        elif action == "broadcast":
            return self.broadcast(message=message, **kwargs)
        elif action == "get_profile":
            return self.get_profile(user_id=to)
        elif action == "get_quota":
            return self.get_message_quota()
        else:
            return {"error": f"Unknown action: {action}"}
    
    def _build_text_message(self, text: str) -> Dict[str, Any]:
        """Build a text message object."""
        return {"type": "text", "text": text}
    
    def _build_flex_message(self, alt_text: str, contents: Dict) -> Dict[str, Any]:
        """Build a Flex Message object."""
        return {
            "type": "flex",
            "altText": alt_text,
            "contents": contents,
        }
    
    def _build_image_message(self, original_url: str, preview_url: Optional[str] = None) -> Dict[str, Any]:
        """Build an image message object."""
        return {
            "type": "image",
            "originalContentUrl": original_url,
            "previewImageUrl": preview_url or original_url,
        }
    
    def _build_sticker_message(self, package_id: str, sticker_id: str) -> Dict[str, Any]:
        """Build a sticker message object."""
        return {
            "type": "sticker",
            "packageId": package_id,
            "stickerId": sticker_id,
        }
    
    def push_message(
        self,
        to: str,
        message: str,
        messages: Optional[List[Dict]] = None,
        notification_disabled: bool = False,
    ) -> Dict[str, Any]:
        """Push a message to a user, group, or room.
        
        Args:
            to: User ID, group ID, or room ID
            message: Text message (ignored if messages is provided)
            messages: List of message objects (max 5)
            notification_disabled: Disable push notification
            
        Returns:
            Send result
        """
        if not to:
            return {"error": "Recipient (to) is required"}
        
        # Normalize target ID
        target_id = to.strip()
        if target_id.lower().startswith("line:"):
            target_id = target_id[5:].strip()
        
        if messages:
            msg_list = messages[:5]  # LINE allows max 5 messages
        elif message:
            msg_list = [self._build_text_message(message)]
        else:
            return {"error": "message or messages is required"}
        
        data = {
            "to": target_id,
            "messages": msg_list,
        }
        
        if notification_disabled:
            data["notificationDisabled"] = True
        
        result = self._request("POST", "/bot/message/push", data)
        
        if "error" not in result:
            return {"success": True, "sent_to": target_id}
        return result
    
    def reply_message(
        self,
        reply_token: str,
        message: str,
        messages: Optional[List[Dict]] = None,
        notification_disabled: bool = False,
    ) -> Dict[str, Any]:
        """Reply to a message using a reply token.
        
        Args:
            reply_token: Reply token from webhook event
            message: Text message (ignored if messages is provided)
            messages: List of message objects (max 5)
            notification_disabled: Disable push notification
            
        Returns:
            Send result
        """
        if not reply_token:
            return {"error": "reply_token is required"}
        
        if messages:
            msg_list = messages[:5]
        elif message:
            msg_list = [self._build_text_message(message)]
        else:
            return {"error": "message or messages is required"}
        
        data = {
            "replyToken": reply_token,
            "messages": msg_list,
        }
        
        if notification_disabled:
            data["notificationDisabled"] = True
        
        result = self._request("POST", "/bot/message/reply", data)
        
        if "error" not in result:
            return {"success": True}
        return result
    
    def multicast(
        self,
        user_ids: List[str],
        message: str,
        messages: Optional[List[Dict]] = None,
        notification_disabled: bool = False,
    ) -> Dict[str, Any]:
        """Send message to multiple users.
        
        Args:
            user_ids: List of user IDs (max 500)
            message: Text message (ignored if messages is provided)
            messages: List of message objects (max 5)
            notification_disabled: Disable push notification
            
        Returns:
            Send result
        """
        if not user_ids:
            return {"error": "user_ids is required"}
        
        if messages:
            msg_list = messages[:5]
        elif message:
            msg_list = [self._build_text_message(message)]
        else:
            return {"error": "message or messages is required"}
        
        data = {
            "to": user_ids[:500],  # LINE allows max 500 users
            "messages": msg_list,
        }
        
        if notification_disabled:
            data["notificationDisabled"] = True
        
        result = self._request("POST", "/bot/message/multicast", data)
        
        if "error" not in result:
            return {"success": True, "sent_to_count": len(user_ids[:500])}
        return result
    
    def broadcast(
        self,
        message: str,
        messages: Optional[List[Dict]] = None,
        notification_disabled: bool = False,
    ) -> Dict[str, Any]:
        """Broadcast message to all followers.
        
        Args:
            message: Text message (ignored if messages is provided)
            messages: List of message objects (max 5)
            notification_disabled: Disable push notification
            
        Returns:
            Send result
        """
        if messages:
            msg_list = messages[:5]
        elif message:
            msg_list = [self._build_text_message(message)]
        else:
            return {"error": "message or messages is required"}
        
        data = {"messages": msg_list}
        
        if notification_disabled:
            data["notificationDisabled"] = True
        
        result = self._request("POST", "/bot/message/broadcast", data)
        
        if "error" not in result:
            return {"success": True, "broadcast": True}
        return result
    
    def get_profile(self, user_id: str) -> Dict[str, Any]:
        """Get user profile.
        
        Args:
            user_id: LINE user ID
            
        Returns:
            User profile
        """
        if not user_id:
            return {"error": "user_id is required"}
        
        result = self._request("GET", f"/bot/profile/{user_id}")
        
        if "error" not in result:
            return {
                "user_id": result.get("userId"),
                "display_name": result.get("displayName"),
                "picture_url": result.get("pictureUrl"),
                "status_message": result.get("statusMessage"),
            }
        return result
    
    def get_message_quota(self) -> Dict[str, Any]:
        """Get message quota information.
        
        Returns:
            Quota information
        """
        result = self._request("GET", "/bot/message/quota")
        
        if "error" not in result:
            return {
                "type": result.get("type"),
                "value": result.get("value"),
            }
        return result
    
    def send_flex_message(
        self,
        to: str,
        alt_text: str,
        contents: Dict,
    ) -> Dict[str, Any]:
        """Send a Flex Message.
        
        Args:
            to: User ID, group ID, or room ID
            alt_text: Alternative text for notifications
            contents: Flex Message contents (bubble or carousel)
            
        Returns:
            Send result
        """
        flex_msg = self._build_flex_message(alt_text, contents)
        return self.push_message(to=to, message="", messages=[flex_msg])
    
    def send_image(
        self,
        to: str,
        image_url: str,
        preview_url: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Send an image message.
        
        Args:
            to: User ID, group ID, or room ID
            image_url: URL of the image
            preview_url: URL of the preview image
            
        Returns:
            Send result
        """
        img_msg = self._build_image_message(image_url, preview_url)
        return self.push_message(to=to, message="", messages=[img_msg])
    
    def verify_signature(self, body: bytes, signature: str) -> bool:
        """Verify webhook signature.
        
        Args:
            body: Request body bytes
            signature: X-Line-Signature header value
            
        Returns:
            True if signature is valid
        """
        if not self.channel_secret:
            logger.warning("LINE_CHANNEL_SECRET not configured, skipping signature verification")
            return True
        
        hash_value = hmac.new(
            self.channel_secret.encode("utf-8"),
            body,
            hashlib.sha256,
        ).digest()
        expected_signature = base64.b64encode(hash_value).decode("utf-8")
        
        return hmac.compare_digest(signature, expected_signature)


def send_line_message(to: str, message: str) -> Dict[str, Any]:
    """Send a LINE message.
    
    Args:
        to: User ID, group ID, or room ID
        message: Message text
        
    Returns:
        Send result
    """
    return LineTool().push_message(to=to, message=message)


def broadcast_line_message(message: str) -> Dict[str, Any]:
    """Broadcast a LINE message to all followers.
    
    Args:
        message: Message text
        
    Returns:
        Send result
    """
    return LineTool().broadcast(message=message)



"""
FCPXML Injector for Final Cut Pro Integration

Manages watch folder delivery and daemon mode for automatic FCPXML injection.
"""

from __future__ import annotations

import json
import os
import signal
import sys
import tempfile
import time
import uuid
from dataclasses import asdict, dataclass
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Callable, Optional

from .commandpost import CommandPostBridge


class JobStatus(str, Enum):
    """Job processing status."""
    PENDING = "pending"
    PROCESSING = "processing"
    INJECTED = "injected"
    FAILED = "failed"


@dataclass
class InjectionJob:
    """Represents an FCPXML injection job."""
    job_id: str
    created_at: str
    instruction: str
    intent_path: Optional[str] = None
    fcpxml_path: Optional[str] = None
    status: JobStatus = JobStatus.PENDING
    error: Optional[str] = None
    completed_at: Optional[str] = None

    def to_dict(self) -> dict:
        """Convert to dictionary."""
        d = asdict(self)
        d["status"] = self.status.value
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "InjectionJob":
        """Create from dictionary."""
        d = d.copy()
        d["status"] = JobStatus(d["status"])
        return cls(**d)


@dataclass
class DaemonState:
    """State of the daemon process."""
    pid: int
    started_at: str
    watch_folder: str
    jobs_processed: int = 0
    last_job_at: Optional[str] = None


class Injector:
    """
    Manages FCPXML injection into Final Cut Pro via CommandPost.
    
    Supports both one-shot injection and daemon mode with watch folder.
    """

    DEFAULT_BASE_DIR = os.path.expanduser("~/.praison/fcp")

    def __init__(
        self,
        base_dir: Optional[str] = None,
        watch_folder: Optional[str] = None,
        commandpost: Optional[CommandPostBridge] = None,
    ):
        """
        Initialize the injector.
        
        Args:
            base_dir: Base directory for jobs and outputs
            watch_folder: Watch folder for CommandPost
            commandpost: CommandPost bridge instance
        """
        self.base_dir = Path(base_dir or self.DEFAULT_BASE_DIR)
        self.jobs_dir = self.base_dir / "jobs"
        self.out_dir = self.base_dir / "out"
        self.watch_folder = Path(watch_folder or self.base_dir / "watch" / "fcpxml")
        self.pid_file = self.base_dir / "daemon.pid"
        self.state_file = self.base_dir / "daemon.state"

        self.commandpost = commandpost or CommandPostBridge(
            watch_folder=str(self.watch_folder)
        )

        self._ensure_directories()

    def _ensure_directories(self) -> None:
        """Create necessary directories."""
        self.jobs_dir.mkdir(parents=True, exist_ok=True)
        self.out_dir.mkdir(parents=True, exist_ok=True)
        self.watch_folder.mkdir(parents=True, exist_ok=True)

    def inject_one_shot(
        self,
        fcpxml_content: str,
        job_id: Optional[str] = None,
        instruction: str = "",
        intent_json: Optional[str] = None,
        delete_after_import: bool = False,
    ) -> tuple[str, str, list[str]]:
        """
        Perform one-shot FCPXML injection.
        
        Args:
            fcpxml_content: FCPXML content to inject
            job_id: Optional job ID (generated if not provided)
            instruction: Original instruction for logging
            intent_json: Optional EditIntent JSON for logging
            delete_after_import: Whether to delete file after import
            
        Returns:
            Tuple of (job_id, fcpxml_path, messages)
        """
        job_id = job_id or str(uuid.uuid4())
        messages = []

        job = InjectionJob(
            job_id=job_id,
            created_at=datetime.utcnow().isoformat(),
            instruction=instruction,
        )

        if intent_json:
            intent_path = self.jobs_dir / f"{job_id}_intent.json"
            intent_path.write_text(intent_json)
            job.intent_path = str(intent_path)
            messages.append(f"Intent saved: {intent_path}")

        fcpxml_path = self._write_fcpxml_atomic(fcpxml_content, job_id)
        job.fcpxml_path = str(fcpxml_path)
        messages.append(f"FCPXML written: {fcpxml_path}")

        watch_path = self._deliver_to_watch_folder(fcpxml_path, job_id)
        messages.append(f"Delivered to watch folder: {watch_path}")

        job.status = JobStatus.INJECTED
        job.completed_at = datetime.utcnow().isoformat()

        job_path = self.jobs_dir / f"{job_id}.json"
        job_path.write_text(json.dumps(job.to_dict(), indent=2))
        messages.append(f"Job logged: {job_path}")

        if delete_after_import:
            try:
                os.remove(fcpxml_path)
                messages.append(f"Cleaned up: {fcpxml_path}")
            except OSError:
                pass

        return job_id, str(watch_path), messages

    def _write_fcpxml_atomic(self, content: str, job_id: str) -> Path:
        """
        Write FCPXML content atomically (write to temp, then rename).
        
        Args:
            content: FCPXML content
            job_id: Job ID for filename
            
        Returns:
            Path to the written file
        """
        final_path = self.out_dir / f"{job_id}.fcpxml"

        fd, tmp_path = tempfile.mkstemp(
            suffix=".fcpxml.tmp",
            dir=str(self.out_dir),
        )
        try:
            with os.fdopen(fd, "w", encoding="utf-8") as f:
                f.write(content)
            os.rename(tmp_path, final_path)
        except Exception:
            if os.path.exists(tmp_path):
                os.remove(tmp_path)
            raise

        return final_path

    def _deliver_to_watch_folder(self, source_path: Path, job_id: str) -> Path:
        """
        Deliver FCPXML to the watch folder with atomic rename.
        
        Args:
            source_path: Source FCPXML file
            job_id: Job ID for filename
            
        Returns:
            Path in watch folder
        """
        watch_path = self.watch_folder / f"{job_id}.fcpxml"

        fd, tmp_path = tempfile.mkstemp(
            suffix=".fcpxml.tmp",
            dir=str(self.watch_folder),
        )
        try:
            with os.fdopen(fd, "w", encoding="utf-8") as f:
                f.write(source_path.read_text())
            os.rename(tmp_path, watch_path)
        except Exception:
            if os.path.exists(tmp_path):
                os.remove(tmp_path)
            raise

        return watch_path

    def get_job(self, job_id: str) -> Optional[InjectionJob]:
        """Get a job by ID."""
        job_path = self.jobs_dir / f"{job_id}.json"
        if not job_path.exists():
            return None
        try:
            return InjectionJob.from_dict(json.loads(job_path.read_text()))
        except (json.JSONDecodeError, KeyError):
            return None

    def list_jobs(self, limit: int = 10) -> list[InjectionJob]:
        """List recent jobs."""
        jobs = []
        for job_file in sorted(
            self.jobs_dir.glob("*.json"),
            key=lambda p: p.stat().st_mtime,
            reverse=True,
        )[:limit]:
            if "_intent" not in job_file.name:
                try:
                    jobs.append(InjectionJob.from_dict(json.loads(job_file.read_text())))
                except (json.JSONDecodeError, KeyError):
                    pass
        return jobs

    def is_daemon_running(self) -> bool:
        """Check if daemon is running."""
        if not self.pid_file.exists():
            return False

        try:
            pid = int(self.pid_file.read_text().strip())
            os.kill(pid, 0)
            return True
        except (ValueError, ProcessLookupError, PermissionError):
            return False

    def get_daemon_state(self) -> Optional[DaemonState]:
        """Get daemon state if running."""
        if not self.is_daemon_running():
            return None

        if not self.state_file.exists():
            return None

        try:
            data = json.loads(self.state_file.read_text())
            return DaemonState(**data)
        except (json.JSONDecodeError, TypeError):
            return None

    def start_daemon(
        self,
        detach: bool = False,
        poll_interval: float = 1.0,
        on_job: Optional[Callable[[InjectionJob], None]] = None,
    ) -> bool:
        """
        Start the daemon process.
        
        Args:
            detach: Whether to detach from terminal
            poll_interval: Seconds between poll cycles
            on_job: Callback when job is processed
            
        Returns:
            True if started successfully
        """
        if self.is_daemon_running():
            return False

        if detach:
            pid = os.fork()
            if pid > 0:
                return True
            os.setsid()
            pid = os.fork()
            if pid > 0:
                sys.exit(0)

        self.pid_file.write_text(str(os.getpid()))

        state = DaemonState(
            pid=os.getpid(),
            started_at=datetime.utcnow().isoformat(),
            watch_folder=str(self.watch_folder),
        )
        self.state_file.write_text(json.dumps(asdict(state)))

        self._running = True

        def handle_signal(signum, frame):
            self._running = False

        signal.signal(signal.SIGTERM, handle_signal)
        signal.signal(signal.SIGINT, handle_signal)

        try:
            self._daemon_loop(poll_interval, on_job, state)
        finally:
            self._cleanup_daemon()

        return True

    def _daemon_loop(
        self,
        poll_interval: float,
        on_job: Optional[Callable[[InjectionJob], None]],
        state: DaemonState,
    ) -> None:
        """Main daemon loop."""
        pending_dir = self.base_dir / "pending"
        pending_dir.mkdir(exist_ok=True)

        while self._running:
            for pending_file in pending_dir.glob("*.json"):
                try:
                    job_data = json.loads(pending_file.read_text())
                    job = InjectionJob.from_dict(job_data)

                    if job.fcpxml_path and os.path.exists(job.fcpxml_path):
                        self._deliver_to_watch_folder(
                            Path(job.fcpxml_path),
                            job.job_id,
                        )
                        job.status = JobStatus.INJECTED
                        job.completed_at = datetime.utcnow().isoformat()
                    else:
                        job.status = JobStatus.FAILED
                        job.error = "FCPXML file not found"

                    job_path = self.jobs_dir / f"{job.job_id}.json"
                    job_path.write_text(json.dumps(job.to_dict(), indent=2))

                    pending_file.unlink()

                    state.jobs_processed += 1
                    state.last_job_at = datetime.utcnow().isoformat()
                    self.state_file.write_text(json.dumps(asdict(state)))

                    if on_job:
                        on_job(job)

                except Exception as e:
                    try:
                        error_path = pending_dir / f"{pending_file.stem}_error.txt"
                        error_path.write_text(str(e))
                        pending_file.unlink()
                    except Exception:
                        pass

            time.sleep(poll_interval)

    def _cleanup_daemon(self) -> None:
        """Clean up daemon files."""
        try:
            self.pid_file.unlink(missing_ok=True)
        except Exception:
            pass

    def stop_daemon(self) -> bool:
        """
        Stop the daemon process.
        
        Returns:
            True if stopped successfully
        """
        if not self.pid_file.exists():
            return False

        try:
            pid = int(self.pid_file.read_text().strip())
            os.kill(pid, signal.SIGTERM)

            for _ in range(50):
                time.sleep(0.1)
                try:
                    os.kill(pid, 0)
                except ProcessLookupError:
                    break

            self.pid_file.unlink(missing_ok=True)
            return True

        except (ValueError, ProcessLookupError, PermissionError):
            self.pid_file.unlink(missing_ok=True)
            return False

    def submit_job(
        self,
        fcpxml_content: str,
        instruction: str = "",
        intent_json: Optional[str] = None,
    ) -> str:
        """
        Submit a job to the daemon queue.
        
        Args:
            fcpxml_content: FCPXML content
            instruction: Original instruction
            intent_json: Optional EditIntent JSON
            
        Returns:
            Job ID
        """
        job_id = str(uuid.uuid4())

        fcpxml_path = self._write_fcpxml_atomic(fcpxml_content, job_id)

        job = InjectionJob(
            job_id=job_id,
            created_at=datetime.utcnow().isoformat(),
            instruction=instruction,
            fcpxml_path=str(fcpxml_path),
        )

        if intent_json:
            intent_path = self.jobs_dir / f"{job_id}_intent.json"
            intent_path.write_text(intent_json)
            job.intent_path = str(intent_path)

        pending_dir = self.base_dir / "pending"
        pending_dir.mkdir(exist_ok=True)

        pending_path = pending_dir / f"{job_id}.json"
        pending_path.write_text(json.dumps(job.to_dict(), indent=2))

        return job_id



"""Tests for ArchiveTool."""

import pytest

from praisonai_tools.recipe_tools.archive_tool import ArchiveTool, archive_create, archive_extract, archive_list


class TestArchiveTool:
    """Unit tests for ArchiveTool."""
    
    @pytest.fixture
    def tool(self):
        return ArchiveTool(verbose=True)
    
    @pytest.mark.unit
    def test_check_dependencies(self, tool):
        """Test dependency checking."""
        deps = tool.check_dependencies()
        assert deps["zipfile"] is True
        assert deps["tarfile"] is True
    
    @pytest.mark.unit
    def test_create_zip(self, tool, sample_folder, temp_dir):
        """Test creating a zip archive."""
        if not sample_folder.exists():
            pytest.skip("Sample folder not found")
        
        output = temp_dir / "test.zip"
        result = tool.create(sample_folder, output, format="zip")
        
        assert result.exists()
        assert result.stat().st_size > 0
    
    @pytest.mark.unit
    def test_create_tar_gz(self, tool, sample_folder, temp_dir):
        """Test creating a tar.gz archive."""
        if not sample_folder.exists():
            pytest.skip("Sample folder not found")
        
        output = temp_dir / "test.tar.gz"
        result = tool.create(sample_folder, output, format="tar.gz")
        
        assert result.exists()
        assert result.stat().st_size > 0
    
    @pytest.mark.unit
    def test_list_zip(self, tool, sample_folder, temp_dir):
        """Test listing zip contents."""
        if not sample_folder.exists():
            pytest.skip("Sample folder not found")
        
        # Create archive first
        archive = temp_dir / "test.zip"
        tool.create(sample_folder, archive, format="zip")
        
        # List contents
        manifest = tool.list(archive)
        
        assert manifest.format == "zip"
        assert manifest.file_count > 0
        assert len(manifest.entries) > 0
    
    @pytest.mark.unit
    def test_extract_zip(self, tool, sample_folder, temp_dir):
        """Test extracting a zip archive."""
        if not sample_folder.exists():
            pytest.skip("Sample folder not found")
        
        # Create archive first
        archive = temp_dir / "test.zip"
        tool.create(sample_folder, archive, format="zip")
        
        # Extract
        extract_dir = temp_dir / "extracted"
        result = tool.extract(archive, extract_dir)
        
        assert result.exists()
        assert result.is_dir()
        # Check some files exist
        assert any(result.rglob("*.txt"))
    
    @pytest.mark.unit
    def test_create_manifest(self, tool, sample_folder, temp_dir):
        """Test creating a manifest."""
        if not sample_folder.exists():
            pytest.skip("Sample folder not found")
        
        output = temp_dir / "manifest.json"
        manifest = tool.create_manifest(sample_folder, output)
        
        assert "files" in manifest
        assert manifest["file_count"] > 0
        assert output.exists()
    
    @pytest.mark.unit
    def test_create_checksums(self, tool, sample_folder, temp_dir):
        """Test creating checksums file."""
        if not sample_folder.exists():
            pytest.skip("Sample folder not found")
        
        output = temp_dir / "checksums.txt"
        content = tool.create_checksums(sample_folder, output)
        
        assert len(content) > 0
        assert output.exists()
        # Check format
        lines = content.split("\n")
        assert all("  " in line for line in lines if line)
    
    @pytest.mark.unit
    def test_excludes(self, tool, temp_dir):
        """Test exclusion patterns."""
        # Create folder with files to exclude
        test_folder = temp_dir / "test_excludes"
        test_folder.mkdir()
        (test_folder / "keep.txt").write_text("keep")
        (test_folder / ".DS_Store").write_text("exclude")
        pycache = test_folder / "__pycache__"
        pycache.mkdir()
        (pycache / "test.pyc").write_text("exclude")
        
        # Create archive
        archive = temp_dir / "test.zip"
        tool.create(test_folder, archive, format="zip")
        
        # List and verify exclusions
        manifest = tool.list(archive)
        names = [e.name for e in manifest.entries]
        
        assert any("keep.txt" in n for n in names)
        assert not any(".DS_Store" in n for n in names)
        assert not any("__pycache__" in n for n in names)


class TestArchiveToolConvenienceFunctions:
    """Test convenience functions."""
    
    @pytest.mark.unit
    def test_archive_create(self, sample_folder, temp_dir):
        """Test archive_create function."""
        if not sample_folder.exists():
            pytest.skip("Sample folder not found")
        
        output = temp_dir / "test.zip"
        result = archive_create(sample_folder, output)
        assert result.exists()
    
    @pytest.mark.unit
    def test_archive_extract(self, sample_folder, temp_dir):
        """Test archive_extract function."""
        if not sample_folder.exists():
            pytest.skip("Sample folder not found")
        
        archive = temp_dir / "test.zip"
        archive_create(sample_folder, archive)
        
        extract_dir = temp_dir / "extracted"
        result = archive_extract(archive, extract_dir)
        assert result.exists()
    
    @pytest.mark.unit
    def test_archive_list(self, sample_folder, temp_dir):
        """Test archive_list function."""
        if not sample_folder.exists():
            pytest.skip("Sample folder not found")
        
        archive = temp_dir / "test.zip"
        archive_create(sample_folder, archive)
        
        manifest = archive_list(archive)
        assert manifest.file_count > 0



"""Apify Tool for PraisonAI Agents.

Run web scraping actors on Apify.

Usage:
    from praisonai_tools import ApifyTool
    
    apify = ApifyTool()
    results = apify.run_actor("apify/web-scraper", {"startUrls": [{"url": "https://example.com"}]})

Environment Variables:
    APIFY_API_TOKEN: Apify API token
"""

import os
import logging
from typing import Any, Dict, List, Optional, Union

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)


class ApifyTool(BaseTool):
    """Tool for Apify web scraping."""
    
    name = "apify"
    description = "Run web scraping actors on Apify."
    
    def __init__(self, api_token: Optional[str] = None):
        self.api_token = api_token or os.getenv("APIFY_API_TOKEN")
        self._client = None
        super().__init__()
    
    @property
    def client(self):
        if self._client is None:
            try:
                from apify_client import ApifyClient
            except ImportError:
                raise ImportError("apify-client not installed. Install with: pip install apify-client")
            if not self.api_token:
                raise ValueError("APIFY_API_TOKEN required")
            self._client = ApifyClient(self.api_token)
        return self._client
    
    def run(
        self,
        action: str = "run_actor",
        actor_id: Optional[str] = None,
        **kwargs
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        if action == "run_actor":
            return self.run_actor(actor_id=actor_id, input_data=kwargs.get("input_data", {}))
        elif action == "get_dataset":
            return self.get_dataset(dataset_id=kwargs.get("dataset_id"))
        return {"error": f"Unknown action: {action}"}
    
    def run_actor(self, actor_id: str, input_data: Dict = None) -> List[Dict[str, Any]]:
        """Run an actor."""
        if not actor_id:
            return [{"error": "actor_id is required"}]
        
        try:
            run = self.client.actor(actor_id).call(run_input=input_data or {})
            dataset = self.client.dataset(run["defaultDatasetId"])
            items = list(dataset.iterate_items())
            return items
        except Exception as e:
            logger.error(f"Apify run_actor error: {e}")
            return [{"error": str(e)}]
    
    def get_dataset(self, dataset_id: str) -> List[Dict[str, Any]]:
        """Get dataset items."""
        if not dataset_id:
            return [{"error": "dataset_id is required"}]
        
        try:
            dataset = self.client.dataset(dataset_id)
            items = list(dataset.iterate_items())
            return items
        except Exception as e:
            logger.error(f"Apify get_dataset error: {e}")
            return [{"error": str(e)}]


def apify_run_actor(actor_id: str, input_data: Dict = None) -> List[Dict[str, Any]]:
    """Run Apify actor."""
    return ApifyTool().run_actor(actor_id=actor_id, input_data=input_data)



"""LLM-based content analysis and edit planning."""

import re
from dataclasses import dataclass, field
from typing import List, Dict, Any, Optional

try:
    from .transcribe import Word, TranscriptResult
except ImportError:
    from transcribe import Word, TranscriptResult


# Common filler words to detect
FILLER_WORDS = {
    "um", "uh", "er", "ah", "like", "you know", "i mean", "basically",
    "actually", "literally", "so", "well", "right", "okay", "ok",
}


@dataclass
class Segment:
    """A segment of video to keep or remove."""
    start: float
    end: float
    action: str  # "keep" or "remove"
    reason: str
    category: str  # "filler", "repetition", "silence", "tangent", "content"
    text: Optional[str] = None
    confidence: float = 1.0
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "start": self.start,
            "end": self.end,
            "action": self.action,
            "reason": self.reason,
            "category": self.category,
            "text": self.text,
            "confidence": self.confidence,
        }


@dataclass
class EditPlan:
    """Complete edit plan for a video."""
    segments: List[Segment] = field(default_factory=list)
    original_duration: float = 0.0
    edited_duration: float = 0.0
    removed_duration: float = 0.0
    removal_summary: Dict[str, float] = field(default_factory=dict)
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "segments": [s.to_dict() for s in self.segments],
            "original_duration": self.original_duration,
            "edited_duration": self.edited_duration,
            "removed_duration": self.removed_duration,
            "removal_summary": self.removal_summary,
        }
    
    def get_keep_segments(self) -> List[Segment]:
        """Get only segments marked for keeping."""
        return [s for s in self.segments if s.action == "keep"]
    
    def get_remove_segments(self) -> List[Segment]:
        """Get only segments marked for removal."""
        return [s for s in self.segments if s.action == "remove"]


def detect_fillers(words: List[Word]) -> List[Segment]:
    """Detect filler words in transcript."""
    segments = []
    
    for word in words:
        text_lower = word.text.lower().strip()
        # Remove punctuation for comparison
        text_clean = re.sub(r'[^\w\s]', '', text_lower)
        
        if text_clean in FILLER_WORDS:
            segments.append(Segment(
                start=word.start,
                end=word.end,
                action="remove",
                reason=f"Filler word: '{word.text}'",
                category="filler",
                text=word.text,
                confidence=0.9,
            ))
    
    return segments


def detect_repetitions(words: List[Word], window: int = 3) -> List[Segment]:
    """Detect repeated words/phrases."""
    segments = []
    
    if len(words) < 2:
        return segments
    
    i = 0
    while i < len(words) - 1:
        # Check for immediate repetition
        curr_text = words[i].text.lower().strip()
        curr_clean = re.sub(r'[^\w]', '', curr_text)
        
        if not curr_clean:
            i += 1
            continue
        
        # Look ahead for repetitions
        j = i + 1
        while j < len(words) and j < i + window:
            next_text = words[j].text.lower().strip()
            next_clean = re.sub(r'[^\w]', '', next_text)
            
            if curr_clean == next_clean and len(curr_clean) > 2:
                # Found repetition - mark the first occurrence for removal
                segments.append(Segment(
                    start=words[i].start,
                    end=words[i].end,
                    action="remove",
                    reason=f"Repeated word: '{words[i].text}'",
                    category="repetition",
                    text=words[i].text,
                    confidence=0.85,
                ))
                break
            j += 1
        i += 1
    
    return segments


def detect_silences(
    words: List[Word],
    duration: float,
    min_silence: float = 1.5,
) -> List[Segment]:
    """Detect long silences between words."""
    segments = []
    
    if not words:
        return segments
    
    # Check silence at start
    if words[0].start > min_silence:
        segments.append(Segment(
            start=0,
            end=words[0].start - 0.2,  # Keep small buffer
            action="remove",
            reason=f"Long silence at start: {words[0].start:.1f}s",
            category="silence",
            confidence=0.95,
        ))
    
    # Check silences between words
    for i in range(len(words) - 1):
        gap = words[i + 1].start - words[i].end
        if gap > min_silence:
            segments.append(Segment(
                start=words[i].end + 0.1,
                end=words[i + 1].start - 0.1,
                action="remove",
                reason=f"Long silence: {gap:.1f}s",
                category="silence",
                confidence=0.95,
            ))
    
    # Check silence at end
    if words and duration - words[-1].end > min_silence:
        segments.append(Segment(
            start=words[-1].end + 0.2,
            end=duration,
            action="remove",
            reason=f"Long silence at end: {duration - words[-1].end:.1f}s",
            category="silence",
            confidence=0.95,
        ))
    
    return segments


def create_edit_plan(
    transcript: TranscriptResult,
    duration: float,
    remove_fillers: bool = True,
    remove_repetitions: bool = True,
    remove_silence: bool = True,
    remove_tangents: bool = False,
    min_silence: float = 1.5,
    use_llm: bool = False,
) -> EditPlan:
    """
    Create an edit plan based on transcript analysis.
    
    Args:
        transcript: Transcription result with word-level timestamps
        duration: Total video duration in seconds
        remove_fillers: Remove filler words (um, uh, like, etc.)
        remove_repetitions: Remove repeated words/phrases
        remove_silence: Remove long silences
        remove_tangents: Use LLM to detect off-topic content
        min_silence: Minimum silence duration to remove (seconds)
        use_llm: Use LLM for advanced analysis
        
    Returns:
        EditPlan with segments to keep/remove
    """
    remove_segments = []
    
    if remove_fillers:
        remove_segments.extend(detect_fillers(transcript.words))
    
    if remove_repetitions:
        remove_segments.extend(detect_repetitions(transcript.words))
    
    if remove_silence:
        remove_segments.extend(detect_silences(
            transcript.words, duration, min_silence
        ))
    
    if remove_tangents and use_llm:
        # LLM-based tangent detection would go here
        # For now, skip as it requires more complex implementation
        pass
    
    # Sort by start time
    remove_segments.sort(key=lambda s: s.start)
    
    # Merge overlapping remove segments
    merged_removes = _merge_overlapping(remove_segments)
    
    # Create keep segments from gaps between removes
    all_segments = _create_keep_segments(merged_removes, duration)
    
    # Calculate statistics
    removed_duration = sum(s.end - s.start for s in merged_removes)
    edited_duration = duration - removed_duration
    
    # Summarize by category
    removal_summary = {}
    for seg in merged_removes:
        cat = seg.category
        removal_summary[cat] = removal_summary.get(cat, 0) + (seg.end - seg.start)
    
    return EditPlan(
        segments=all_segments,
        original_duration=duration,
        edited_duration=edited_duration,
        removed_duration=removed_duration,
        removal_summary=removal_summary,
    )


def _merge_overlapping(segments: List[Segment]) -> List[Segment]:
    """Merge overlapping segments."""
    if not segments:
        return []
    
    merged = [segments[0]]
    
    for seg in segments[1:]:
        last = merged[-1]
        # If overlapping or adjacent (within 0.1s)
        if seg.start <= last.end + 0.1:
            # Extend the last segment
            merged[-1] = Segment(
                start=last.start,
                end=max(last.end, seg.end),
                action="remove",
                reason=f"{last.reason}; {seg.reason}",
                category=last.category if last.category == seg.category else "mixed",
                confidence=min(last.confidence, seg.confidence),
            )
        else:
            merged.append(seg)
    
    return merged


def _create_keep_segments(
    remove_segments: List[Segment],
    duration: float,
) -> List[Segment]:
    """Create keep segments from gaps between remove segments."""
    all_segments = []
    current_time = 0.0
    
    for remove_seg in remove_segments:
        # Add keep segment before this remove
        if remove_seg.start > current_time + 0.05:
            all_segments.append(Segment(
                start=current_time,
                end=remove_seg.start,
                action="keep",
                reason="Content",
                category="content",
                confidence=1.0,
            ))
        
        # Add the remove segment
        all_segments.append(remove_seg)
        current_time = remove_seg.end
    
    # Add final keep segment
    if current_time < duration - 0.05:
        all_segments.append(Segment(
            start=current_time,
            end=duration,
            action="keep",
            reason="Content",
            category="content",
            confidence=1.0,
        ))
    
    return all_segments



"""
Image Tool - Image operations via Pillow and ImageMagick.

Provides:
- Probing image files for metadata
- Optimizing images (compression, format conversion)
- Resizing images
- Creating thumbnails
- Creating montages/contact sheets
"""

import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import sys
import os
_dir = os.path.dirname(os.path.abspath(__file__))
if _dir not in sys.path:
    sys.path.insert(0, _dir)

try:
    from .base import RecipeToolBase
except ImportError:
    from base import RecipeToolBase

logger = logging.getLogger(__name__)

# Try to import Pillow
try:
    from PIL import Image, ImageOps
    PILLOW_AVAILABLE = True
except ImportError:
    PILLOW_AVAILABLE = False
    Image = None
    ImageOps = None


@dataclass
class ImageProbeResult:
    """Result of probing an image file."""
    path: str
    format: str
    width: int
    height: int
    mode: str  # RGB, RGBA, L, etc.
    file_size: int
    has_alpha: bool = False
    is_animated: bool = False
    frame_count: int = 1
    dpi: Optional[Tuple[int, int]] = None
    exif: Dict = field(default_factory=dict)
    
    def to_dict(self) -> Dict:
        """Convert to dictionary."""
        return {
            "path": self.path,
            "format": self.format,
            "width": self.width,
            "height": self.height,
            "mode": self.mode,
            "file_size": self.file_size,
            "has_alpha": self.has_alpha,
            "is_animated": self.is_animated,
            "frame_count": self.frame_count,
            "dpi": self.dpi,
        }


class ImageTool(RecipeToolBase):
    """
    Image operations tool using Pillow and ImageMagick.
    
    Provides image probing, optimization, resizing, and montage creation.
    """
    
    name = "image_tool"
    description = "Image operations via Pillow/ImageMagick"
    
    def check_dependencies(self) -> Dict[str, bool]:
        """Check for Pillow and ImageMagick."""
        return {
            "pillow": PILLOW_AVAILABLE,
            "convert": self._check_binary("convert"),
            "montage": self._check_binary("montage"),
            "identify": self._check_binary("identify"),
        }
    
    def _require_pillow(self):
        """Ensure Pillow is available."""
        if not PILLOW_AVAILABLE:
            raise ImportError("Pillow is required. Install with: pip install Pillow")
    
    def probe(self, path: Union[str, Path]) -> ImageProbeResult:
        """
        Probe an image file for metadata.
        
        Args:
            path: Path to image file
            
        Returns:
            ImageProbeResult with file metadata
        """
        self._require_pillow()
        path = Path(path)
        
        if not path.exists():
            raise FileNotFoundError(f"Image not found: {path}")
        
        with Image.open(path) as img:
            # Basic info
            width, height = img.size
            format_name = img.format or path.suffix[1:].upper()
            mode = img.mode
            
            # Check for alpha
            has_alpha = mode in ("RGBA", "LA", "PA") or "transparency" in img.info
            
            # Check for animation
            is_animated = getattr(img, "is_animated", False)
            frame_count = getattr(img, "n_frames", 1)
            
            # Get DPI
            dpi = img.info.get("dpi")
            
            # Get EXIF data
            exif = {}
            try:
                exif_data = img._getexif()
                if exif_data:
                    exif = {k: str(v) for k, v in exif_data.items() if isinstance(v, (str, int, float))}
            except (AttributeError, Exception):
                pass
        
        return ImageProbeResult(
            path=str(path),
            format=format_name,
            width=width,
            height=height,
            mode=mode,
            file_size=path.stat().st_size,
            has_alpha=has_alpha,
            is_animated=is_animated,
            frame_count=frame_count,
            dpi=dpi,
            exif=exif,
        )
    
    def optimize(
        self,
        input_path: Union[str, Path],
        output_path: Optional[Union[str, Path]] = None,
        quality: int = 85,
        max_size: Optional[Tuple[int, int]] = None,
        format: Optional[str] = None,
    ) -> Path:
        """
        Optimize an image for web/storage.
        
        Args:
            input_path: Input image file
            output_path: Output file (auto-generated if not provided)
            quality: JPEG/WebP quality (1-100)
            max_size: Maximum dimensions (width, height)
            format: Output format (jpg, png, webp)
            
        Returns:
            Path to optimized image
        """
        self._require_pillow()
        input_path = Path(input_path)
        
        if not input_path.exists():
            raise FileNotFoundError(f"Image not found: {input_path}")
        
        # Determine output format
        if format is None:
            format = input_path.suffix[1:].lower()
        format = format.lower()
        if format == "jpeg":
            format = "jpg"
        
        if output_path is None:
            output_path = self._generate_output_path(input_path, suffix="_optimized", extension=format)
        else:
            output_path = Path(output_path)
            self._ensure_output_dir(output_path.parent)
        
        with Image.open(input_path) as img:
            # Convert mode if needed
            if format in ("jpg", "jpeg") and img.mode in ("RGBA", "LA", "P"):
                img = img.convert("RGB")
            
            # Resize if max_size specified
            if max_size:
                img.thumbnail(max_size, Image.Resampling.LANCZOS)
            
            # Auto-orient based on EXIF
            img = ImageOps.exif_transpose(img)
            
            # Save with optimization
            save_kwargs = {"optimize": True}
            if format in ("jpg", "jpeg"):
                save_kwargs["quality"] = quality
                save_kwargs["progressive"] = True
            elif format == "webp":
                save_kwargs["quality"] = quality
            elif format == "png":
                save_kwargs["compress_level"] = 9
            
            img.save(output_path, **save_kwargs)
        
        return output_path
    
    def resize(
        self,
        input_path: Union[str, Path],
        output_path: Optional[Union[str, Path]] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        scale: Optional[float] = None,
        keep_aspect: bool = True,
    ) -> Path:
        """
        Resize an image.
        
        Args:
            input_path: Input image file
            output_path: Output file (auto-generated if not provided)
            width: Target width
            height: Target height
            scale: Scale factor (e.g., 0.5 for half size)
            keep_aspect: Maintain aspect ratio
            
        Returns:
            Path to resized image
        """
        self._require_pillow()
        input_path = Path(input_path)
        
        if not input_path.exists():
            raise FileNotFoundError(f"Image not found: {input_path}")
        
        if output_path is None:
            suffix = f"_{width}x{height}" if width and height else "_resized"
            output_path = self._generate_output_path(input_path, suffix=suffix)
        else:
            output_path = Path(output_path)
            self._ensure_output_dir(output_path.parent)
        
        with Image.open(input_path) as img:
            orig_width, orig_height = img.size
            
            # Calculate new dimensions
            if scale:
                new_width = int(orig_width * scale)
                new_height = int(orig_height * scale)
            elif width and height:
                if keep_aspect:
                    img.thumbnail((width, height), Image.Resampling.LANCZOS)
                    img.save(output_path)
                    return output_path
                else:
                    new_width, new_height = width, height
            elif width:
                ratio = width / orig_width
                new_width = width
                new_height = int(orig_height * ratio) if keep_aspect else orig_height
            elif height:
                ratio = height / orig_height
                new_height = height
                new_width = int(orig_width * ratio) if keep_aspect else orig_width
            else:
                new_width, new_height = orig_width, orig_height
            
            resized = img.resize((new_width, new_height), Image.Resampling.LANCZOS)
            resized.save(output_path)
        
        return output_path
    
    def thumbnail(
        self,
        input_path: Union[str, Path],
        output_path: Optional[Union[str, Path]] = None,
        size: Tuple[int, int] = (256, 256),
        format: str = "jpg",
        quality: int = 85,
    ) -> Path:
        """
        Create a thumbnail from an image.
        
        Args:
            input_path: Input image file
            output_path: Output file (auto-generated if not provided)
            size: Maximum thumbnail dimensions
            format: Output format
            quality: JPEG quality
            
        Returns:
            Path to thumbnail
        """
        self._require_pillow()
        input_path = Path(input_path)
        
        if not input_path.exists():
            raise FileNotFoundError(f"Image not found: {input_path}")
        
        if output_path is None:
            output_path = self._generate_output_path(input_path, suffix="_thumb", extension=format)
        else:
            output_path = Path(output_path)
            self._ensure_output_dir(output_path.parent)
        
        with Image.open(input_path) as img:
            # Convert mode if needed
            if format in ("jpg", "jpeg") and img.mode in ("RGBA", "LA", "P"):
                img = img.convert("RGB")
            
            # Auto-orient
            img = ImageOps.exif_transpose(img)
            
            # Create thumbnail
            img.thumbnail(size, Image.Resampling.LANCZOS)
            
            # Save
            save_kwargs = {}
            if format in ("jpg", "jpeg"):
                save_kwargs["quality"] = quality
            
            img.save(output_path, **save_kwargs)
        
        return output_path
    
    def montage(
        self,
        input_paths: List[Union[str, Path]],
        output_path: Union[str, Path],
        columns: int = 4,
        tile_size: Tuple[int, int] = (200, 200),
        background: str = "white",
        border: int = 2,
    ) -> Path:
        """
        Create a montage/contact sheet from multiple images.
        
        Uses ImageMagick if available, falls back to Pillow.
        
        Args:
            input_paths: List of input image paths
            output_path: Output file path
            columns: Number of columns
            tile_size: Size of each tile
            background: Background color
            border: Border width between tiles
            
        Returns:
            Path to montage image
        """
        output_path = Path(output_path)
        self._ensure_output_dir(output_path.parent)
        
        # Try ImageMagick first (better quality for large montages)
        deps = self.check_dependencies()
        if deps.get("montage"):
            return self._montage_imagemagick(
                input_paths, output_path, columns, tile_size, background, border
            )
        
        # Fall back to Pillow
        self._require_pillow()
        return self._montage_pillow(
            input_paths, output_path, columns, tile_size, background, border
        )
    
    def _montage_imagemagick(
        self,
        input_paths: List[Union[str, Path]],
        output_path: Path,
        columns: int,
        tile_size: Tuple[int, int],
        background: str,
        border: int,
    ) -> Path:
        """Create montage using ImageMagick."""
        cmd = [
            "montage",
            "-geometry", f"{tile_size[0]}x{tile_size[1]}+{border}+{border}",
            "-tile", f"{columns}x",
            "-background", background,
        ]
        
        for p in input_paths:
            cmd.append(str(p))
        
        cmd.append(str(output_path))
        
        result = self._run_command(cmd)
        if result.returncode != 0:
            raise RuntimeError(f"montage failed: {result.stderr}")
        
        return output_path
    
    def _montage_pillow(
        self,
        input_paths: List[Union[str, Path]],
        output_path: Path,
        columns: int,
        tile_size: Tuple[int, int],
        background: str,
        border: int,
    ) -> Path:
        """Create montage using Pillow."""
        n_images = len(input_paths)
        rows = (n_images + columns - 1) // columns
        
        # Calculate canvas size
        canvas_width = columns * (tile_size[0] + border * 2)
        canvas_height = rows * (tile_size[1] + border * 2)
        
        # Create canvas
        canvas = Image.new("RGB", (canvas_width, canvas_height), background)
        
        for i, path in enumerate(input_paths):
            row = i // columns
            col = i % columns
            
            x = col * (tile_size[0] + border * 2) + border
            y = row * (tile_size[1] + border * 2) + border
            
            try:
                with Image.open(path) as img:
                    # Convert to RGB if needed
                    if img.mode in ("RGBA", "LA", "P"):
                        img = img.convert("RGB")
                    
                    # Create thumbnail
                    img.thumbnail(tile_size, Image.Resampling.LANCZOS)
                    
                    # Center in tile
                    offset_x = (tile_size[0] - img.width) // 2
                    offset_y = (tile_size[1] - img.height) // 2
                    
                    canvas.paste(img, (x + offset_x, y + offset_y))
            except Exception as e:
                logger.warning(f"Failed to process {path}: {e}")
        
        canvas.save(output_path)
        return output_path
    
    def batch_resize(
        self,
        input_dir: Union[str, Path],
        output_dir: Optional[Union[str, Path]] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        extensions: List[str] = None,
    ) -> List[Path]:
        """
        Batch resize images in a directory.
        
        Args:
            input_dir: Input directory
            output_dir: Output directory
            width: Target width
            height: Target height
            extensions: File extensions to process
            
        Returns:
            List of paths to resized images
        """
        input_dir = Path(input_dir)
        
        if not input_dir.is_dir():
            raise NotADirectoryError(f"Not a directory: {input_dir}")
        
        if output_dir is None:
            output_dir = input_dir / "resized"
        else:
            output_dir = Path(output_dir)
        
        self._ensure_output_dir(output_dir)
        
        if extensions is None:
            extensions = ["jpg", "jpeg", "png", "webp", "gif", "bmp"]
        
        results = []
        for ext in extensions:
            for img_path in input_dir.glob(f"*.{ext}"):
                try:
                    output_path = output_dir / img_path.name
                    self.resize(img_path, output_path, width=width, height=height)
                    results.append(output_path)
                except Exception as e:
                    logger.warning(f"Failed to resize {img_path}: {e}")
        
        return results


# Convenience functions
def image_optimize(
    input_path: Union[str, Path],
    output_path: Optional[Union[str, Path]] = None,
    quality: int = 85,
    max_size: Optional[Tuple[int, int]] = None,
    verbose: bool = False,
) -> Path:
    """Optimize an image."""
    return ImageTool(verbose=verbose).optimize(input_path, output_path, quality, max_size)


def image_resize(
    input_path: Union[str, Path],
    output_path: Optional[Union[str, Path]] = None,
    width: Optional[int] = None,
    height: Optional[int] = None,
    verbose: bool = False,
) -> Path:
    """Resize an image."""
    return ImageTool(verbose=verbose).resize(input_path, output_path, width, height)


def image_thumbnail(
    input_path: Union[str, Path],
    output_path: Optional[Union[str, Path]] = None,
    size: Tuple[int, int] = (256, 256),
    verbose: bool = False,
) -> Path:
    """Create a thumbnail."""
    return ImageTool(verbose=verbose).thumbnail(input_path, output_path, size)


def image_montage(
    input_paths: List[Union[str, Path]],
    output_path: Union[str, Path],
    columns: int = 4,
    verbose: bool = False,
) -> Path:
    """Create a montage."""
    return ImageTool(verbose=verbose).montage(input_paths, output_path, columns)



"""Local FileSystem Tool for PraisonAI Agents.

Local file system operations with directory traversal.

Usage:
    from praisonai_tools import LocalFileSystemTool
    
    fs = LocalFileSystemTool()
    files = fs.list_directory("/path/to/dir")
"""

import os
import logging
from typing import Any, Dict, List, Optional, Union

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)


class LocalFileSystemTool(BaseTool):
    """Tool for local file system operations."""
    
    name = "local_filesystem"
    description = "Local file system operations with directory traversal."
    
    def __init__(self, base_path: str = None):
        self.base_path = base_path or os.getcwd()
        super().__init__()
    
    def run(
        self,
        action: str = "list",
        path: Optional[str] = None,
        **kwargs
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        action = action.lower().replace("-", "_")
        
        if action == "list":
            return self.list_directory(path=path)
        elif action == "read":
            return self.read_file(path=path)
        elif action == "write":
            return self.write_file(path=path, content=kwargs.get("content"))
        elif action == "exists":
            return self.exists(path=path)
        elif action == "mkdir":
            return self.mkdir(path=path)
        elif action == "delete":
            return self.delete(path=path)
        elif action == "tree":
            return self.tree(path=path, **kwargs)
        else:
            return {"error": f"Unknown action: {action}"}
    
    def _resolve_path(self, path: str) -> str:
        if os.path.isabs(path):
            return path
        return os.path.join(self.base_path, path)
    
    def list_directory(self, path: str = ".") -> List[Dict[str, Any]]:
        """List directory contents."""
        full_path = self._resolve_path(path or ".")
        
        try:
            items = []
            for name in os.listdir(full_path):
                item_path = os.path.join(full_path, name)
                stat = os.stat(item_path)
                items.append({
                    "name": name,
                    "type": "directory" if os.path.isdir(item_path) else "file",
                    "size": stat.st_size,
                    "modified": stat.st_mtime,
                })
            return items
        except Exception as e:
            logger.error(f"LocalFileSystem list error: {e}")
            return [{"error": str(e)}]
    
    def read_file(self, path: str) -> Dict[str, Any]:
        """Read file content."""
        if not path:
            return {"error": "path is required"}
        
        full_path = self._resolve_path(path)
        
        try:
            with open(full_path, "r", encoding="utf-8") as f:
                content = f.read()
            return {"path": path, "content": content}
        except Exception as e:
            logger.error(f"LocalFileSystem read error: {e}")
            return {"error": str(e)}
    
    def write_file(self, path: str, content: str) -> Dict[str, Any]:
        """Write file content."""
        if not path or content is None:
            return {"error": "path and content are required"}
        
        full_path = self._resolve_path(path)
        
        try:
            os.makedirs(os.path.dirname(full_path), exist_ok=True)
            with open(full_path, "w", encoding="utf-8") as f:
                f.write(content)
            return {"success": True, "path": path}
        except Exception as e:
            logger.error(f"LocalFileSystem write error: {e}")
            return {"error": str(e)}
    
    def exists(self, path: str) -> Dict[str, Any]:
        """Check if path exists."""
        if not path:
            return {"error": "path is required"}
        
        full_path = self._resolve_path(path)
        return {"exists": os.path.exists(full_path), "is_file": os.path.isfile(full_path), "is_dir": os.path.isdir(full_path)}
    
    def mkdir(self, path: str) -> Dict[str, Any]:
        """Create directory."""
        if not path:
            return {"error": "path is required"}
        
        full_path = self._resolve_path(path)
        
        try:
            os.makedirs(full_path, exist_ok=True)
            return {"success": True, "path": path}
        except Exception as e:
            logger.error(f"LocalFileSystem mkdir error: {e}")
            return {"error": str(e)}
    
    def delete(self, path: str) -> Dict[str, Any]:
        """Delete file or empty directory."""
        if not path:
            return {"error": "path is required"}
        
        full_path = self._resolve_path(path)
        
        try:
            if os.path.isfile(full_path):
                os.remove(full_path)
            elif os.path.isdir(full_path):
                os.rmdir(full_path)
            return {"success": True}
        except Exception as e:
            logger.error(f"LocalFileSystem delete error: {e}")
            return {"error": str(e)}
    
    def tree(self, path: str = ".", max_depth: int = 3) -> List[Dict[str, Any]]:
        """Get directory tree."""
        full_path = self._resolve_path(path or ".")
        
        def _walk(p, depth):
            if depth > max_depth:
                return []
            items = []
            try:
                for name in os.listdir(p):
                    item_path = os.path.join(p, name)
                    is_dir = os.path.isdir(item_path)
                    item = {"name": name, "type": "directory" if is_dir else "file"}
                    if is_dir:
                        item["children"] = _walk(item_path, depth + 1)
                    items.append(item)
            except PermissionError:
                pass
            return items
        
        return _walk(full_path, 1)


def list_local_directory(path: str = ".") -> List[Dict[str, Any]]:
    """List local directory."""
    return LocalFileSystemTool().list_directory(path=path)



"""PGVector Tool for PraisonAI Agents.

Vector operations using PostgreSQL with pgvector.

Usage:
    from praisonai_tools import PGVectorTool
    
    pgv = PGVectorTool()
    results = pgv.search(table="documents", vector=[...])

Environment Variables:
    PGVECTOR_HOST: PostgreSQL host
    PGVECTOR_PORT: PostgreSQL port
    PGVECTOR_USER: Username
    PGVECTOR_PASSWORD: Password
    PGVECTOR_DATABASE: Database name
"""

import os
import logging
from typing import Any, Dict, List, Optional, Union

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)


class PGVectorTool(BaseTool):
    """Tool for PGVector operations."""
    
    name = "pgvector"
    description = "Vector operations using PostgreSQL with pgvector."
    
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        database: Optional[str] = None,
    ):
        self.host = host or os.getenv("PGVECTOR_HOST", "localhost")
        self.port = port or int(os.getenv("PGVECTOR_PORT", "5432"))
        self.user = user or os.getenv("PGVECTOR_USER")
        self.password = password or os.getenv("PGVECTOR_PASSWORD")
        self.database = database or os.getenv("PGVECTOR_DATABASE")
        self._conn = None
        super().__init__()
    
    @property
    def conn(self):
        if self._conn is None:
            try:
                import psycopg2
            except ImportError:
                raise ImportError("psycopg2 not installed")
            self._conn = psycopg2.connect(
                host=self.host,
                port=self.port,
                user=self.user,
                password=self.password,
                database=self.database,
            )
        return self._conn
    
    def run(
        self,
        action: str = "search",
        table: Optional[str] = None,
        vector: Optional[List[float]] = None,
        **kwargs
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        action = action.lower().replace("-", "_")
        
        if action == "search":
            return self.search(table=table, vector=vector, **kwargs)
        elif action == "insert":
            return self.insert(table=table, vector=vector, **kwargs)
        else:
            return {"error": f"Unknown action: {action}"}
    
    def search(self, table: str, vector: List[float], limit: int = 10, column: str = "embedding") -> List[Dict[str, Any]]:
        """Search by vector similarity."""
        if not table or not vector:
            return [{"error": "table and vector are required"}]
        
        try:
            with self.conn.cursor() as cursor:
                vector_str = "[" + ",".join(map(str, vector)) + "]"
                sql = f"SELECT *, {column} <-> %s::vector AS distance FROM {table} ORDER BY distance LIMIT %s"
                cursor.execute(sql, (vector_str, limit))
                columns = [desc[0] for desc in cursor.description]
                return [dict(zip(columns, row)) for row in cursor.fetchall()]
        except Exception as e:
            logger.error(f"PGVector search error: {e}")
            return [{"error": str(e)}]
    
    def insert(self, table: str, vector: List[float], data: Dict = None, column: str = "embedding") -> Dict[str, Any]:
        """Insert vector."""
        if not table or not vector:
            return {"error": "table and vector are required"}
        
        try:
            with self.conn.cursor() as cursor:
                vector_str = "[" + ",".join(map(str, vector)) + "]"
                if data:
                    cols = list(data.keys()) + [column]
                    vals = list(data.values()) + [vector_str]
                    placeholders = ", ".join(["%s"] * len(vals))
                    sql = f"INSERT INTO {table} ({', '.join(cols)}) VALUES ({placeholders})"
                else:
                    sql = f"INSERT INTO {table} ({column}) VALUES (%s::vector)"
                    vals = [vector_str]
                cursor.execute(sql, vals)
                self.conn.commit()
                return {"success": True}
        except Exception as e:
            logger.error(f"PGVector insert error: {e}")
            return {"error": str(e)}


def pgvector_search(table: str, vector: List[float], limit: int = 10) -> List[Dict[str, Any]]:
    """Search PGVector."""
    return PGVectorTool().search(table=table, vector=vector, limit=limit)



"""Slack Tool for PraisonAI Agents.

Provides Slack messaging capabilities - send messages, read channels, manage threads.

Usage:
    from praisonai_tools import SlackTool
    
    slack = SlackTool()  # Uses SLACK_TOKEN env var
    
    # Send message
    slack.send_message(channel="#general", text="Hello from AI!")
    
    # List channels
    channels = slack.list_channels()
    
    # Get channel history
    messages = slack.get_history(channel="C01234567", limit=10)

Environment Variables:
    SLACK_TOKEN: Slack Bot User OAuth Token (xoxb-...)
    
Required Slack Bot Scopes:
    - chat:write (send messages)
    - channels:read (list channels)
    - channels:history (read channel history)
    - groups:read (list private channels)
    - groups:history (read private channel history)
"""

import os
import logging
from typing import Any, Dict, List, Optional, Union

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)


class SlackTool(BaseTool):
    """Tool for interacting with Slack workspaces.
    
    Send messages, read channel history, list channels, and reply to threads.
    
    Attributes:
        name: Tool identifier
        description: Tool description for LLM
        token: Slack Bot User OAuth Token
        markdown: Enable Slack markdown formatting
    """
    
    name = "slack"
    description = "Send messages to Slack channels, read channel history, list channels, and reply to threads."
    
    def __init__(
        self,
        token: Optional[str] = None,
        markdown: bool = True,
        default_channel: Optional[str] = None,
    ):
        """Initialize SlackTool.
        
        Args:
            token: Slack Bot User OAuth Token (or use SLACK_TOKEN env var)
            markdown: Enable Slack markdown formatting (default: True)
            default_channel: Default channel for messages
        """
        self.token = token or os.getenv("SLACK_TOKEN")
        self.markdown = markdown
        self.default_channel = default_channel
        self._client = None
        
        super().__init__()
    
    @property
    def client(self):
        """Lazy-load Slack client."""
        if self._client is None:
            if not self.token:
                raise ValueError("SLACK_TOKEN not configured. Set SLACK_TOKEN environment variable.")
            try:
                from slack_sdk import WebClient
            except ImportError:
                raise ImportError(
                    "slack_sdk not installed. Install with: pip install slack-sdk"
                )
            self._client = WebClient(token=self.token)
        return self._client
    
    def run(
        self,
        action: str = "send",
        channel: Optional[str] = None,
        text: Optional[str] = None,
        thread_ts: Optional[str] = None,
        limit: int = 10,
        **kwargs
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        """Execute Slack action.
        
        Args:
            action: "send", "list_channels", "history", "reply"
            channel: Channel ID or name (e.g., "#general" or "C01234567")
            text: Message text to send
            thread_ts: Thread timestamp for replies
            limit: Max messages for history
            
        Returns:
            Action result
        """
        action = action.lower()
        channel = channel or self.default_channel
        
        if action == "send":
            return self.send_message(channel=channel, text=text)
        elif action == "reply":
            return self.reply_to_thread(channel=channel, text=text, thread_ts=thread_ts)
        elif action == "list_channels":
            return self.list_channels()
        elif action == "history":
            return self.get_history(channel=channel, limit=limit)
        elif action == "users":
            return self.list_users()
        else:
            return f"Unknown action: {action}. Use 'send', 'reply', 'list_channels', 'history', or 'users'."
    
    def send_message(
        self,
        channel: str,
        text: str,
        blocks: Optional[List[Dict]] = None,
    ) -> Dict[str, Any]:
        """Send a message to a Slack channel.
        
        Args:
            channel: Channel ID or name
            text: Message text
            blocks: Optional Block Kit blocks for rich formatting
            
        Returns:
            API response with message details
        """
        if not channel:
            return {"error": "Channel is required"}
        if not text:
            return {"error": "Text is required"}
        
        try:
            from slack_sdk.errors import SlackApiError
        except ImportError:
            return {"error": "slack_sdk not installed"}
        
        try:
            kwargs = {
                "channel": channel,
                "text": text,
                "mrkdwn": self.markdown,
            }
            if blocks:
                kwargs["blocks"] = blocks
            
            response = self.client.chat_postMessage(**kwargs)
            
            return {
                "success": True,
                "channel": response["channel"],
                "ts": response["ts"],
                "message": response.get("message", {}).get("text", text),
            }
        except SlackApiError as e:
            logger.error(f"Slack API error: {e}")
            return {"error": str(e.response["error"])}
        except Exception as e:
            logger.error(f"Slack error: {e}")
            return {"error": str(e)}
    
    def reply_to_thread(
        self,
        channel: str,
        text: str,
        thread_ts: str,
    ) -> Dict[str, Any]:
        """Reply to a thread in a Slack channel.
        
        Args:
            channel: Channel ID
            text: Reply text
            thread_ts: Parent message timestamp
            
        Returns:
            API response
        """
        if not channel:
            return {"error": "Channel is required"}
        if not text:
            return {"error": "Text is required"}
        if not thread_ts:
            return {"error": "thread_ts is required for replies"}
        
        try:
            from slack_sdk.errors import SlackApiError
        except ImportError:
            return {"error": "slack_sdk not installed"}
        
        try:
            response = self.client.chat_postMessage(
                channel=channel,
                text=text,
                thread_ts=thread_ts,
                mrkdwn=self.markdown,
            )
            
            return {
                "success": True,
                "channel": response["channel"],
                "ts": response["ts"],
                "thread_ts": thread_ts,
            }
        except SlackApiError as e:
            logger.error(f"Slack API error: {e}")
            return {"error": str(e.response["error"])}
        except Exception as e:
            logger.error(f"Slack error: {e}")
            return {"error": str(e)}
    
    def list_channels(
        self,
        include_private: bool = False,
        limit: int = 100,
    ) -> List[Dict[str, Any]]:
        """List channels in the workspace.
        
        Args:
            include_private: Include private channels
            limit: Max channels to return
            
        Returns:
            List of channel info dicts
        """
        try:
            from slack_sdk.errors import SlackApiError
        except ImportError:
            return [{"error": "slack_sdk not installed"}]
        
        try:
            types = "public_channel,private_channel" if include_private else "public_channel"
            response = self.client.conversations_list(types=types, limit=limit)
            
            channels = []
            for ch in response.get("channels", []):
                channels.append({
                    "id": ch["id"],
                    "name": ch["name"],
                    "is_private": ch.get("is_private", False),
                    "num_members": ch.get("num_members", 0),
                    "topic": ch.get("topic", {}).get("value", ""),
                })
            
            return channels
        except SlackApiError as e:
            logger.error(f"Slack API error: {e}")
            return [{"error": str(e.response["error"])}]
        except Exception as e:
            logger.error(f"Slack error: {e}")
            return [{"error": str(e)}]
    
    def get_history(
        self,
        channel: str,
        limit: int = 10,
        oldest: Optional[str] = None,
        latest: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """Get message history from a channel.
        
        Args:
            channel: Channel ID
            limit: Max messages to return
            oldest: Only messages after this timestamp
            latest: Only messages before this timestamp
            
        Returns:
            List of message dicts
        """
        if not channel:
            return [{"error": "Channel is required"}]
        
        try:
            from slack_sdk.errors import SlackApiError
        except ImportError:
            return [{"error": "slack_sdk not installed"}]
        
        try:
            kwargs = {"channel": channel, "limit": limit}
            if oldest:
                kwargs["oldest"] = oldest
            if latest:
                kwargs["latest"] = latest
            
            response = self.client.conversations_history(**kwargs)
            
            messages = []
            for msg in response.get("messages", []):
                messages.append({
                    "text": msg.get("text", ""),
                    "user": msg.get("user", "bot") if msg.get("subtype") != "bot_message" else "bot",
                    "ts": msg.get("ts", ""),
                    "type": msg.get("subtype", "message"),
                    "thread_ts": msg.get("thread_ts"),
                    "reply_count": msg.get("reply_count", 0),
                })
            
            return messages
        except SlackApiError as e:
            logger.error(f"Slack API error: {e}")
            return [{"error": str(e.response["error"])}]
        except Exception as e:
            logger.error(f"Slack error: {e}")
            return [{"error": str(e)}]
    
    def list_users(self, limit: int = 100) -> List[Dict[str, Any]]:
        """List users in the workspace.
        
        Args:
            limit: Max users to return
            
        Returns:
            List of user info dicts
        """
        try:
            from slack_sdk.errors import SlackApiError
        except ImportError:
            return [{"error": "slack_sdk not installed"}]
        
        try:
            response = self.client.users_list(limit=limit)
            
            users = []
            for user in response.get("members", []):
                if not user.get("deleted") and not user.get("is_bot"):
                    users.append({
                        "id": user["id"],
                        "name": user.get("name", ""),
                        "real_name": user.get("real_name", ""),
                        "email": user.get("profile", {}).get("email", ""),
                    })
            
            return users
        except SlackApiError as e:
            logger.error(f"Slack API error: {e}")
            return [{"error": str(e.response["error"])}]
        except Exception as e:
            logger.error(f"Slack error: {e}")
            return [{"error": str(e)}]


# Convenience functions
def send_slack_message(
    channel: str,
    text: str,
    thread_ts: Optional[str] = None,
) -> Dict[str, Any]:
    """Send a Slack message using environment credentials.
    
    Args:
        channel: Channel ID or name
        text: Message text
        thread_ts: Optional thread to reply to
        
    Returns:
        API response
    """
    tool = SlackTool()
    if thread_ts:
        return tool.reply_to_thread(channel=channel, text=text, thread_ts=thread_ts)
    return tool.send_message(channel=channel, text=text)


def get_slack_channels() -> List[Dict[str, Any]]:
    """List Slack channels using environment credentials."""
    tool = SlackTool()
    return tool.list_channels()


def get_slack_history(channel: str, limit: int = 10) -> List[Dict[str, Any]]:
    """Get Slack channel history using environment credentials."""
    tool = SlackTool()
    return tool.get_history(channel=channel, limit=limit)



"""WhatsApp Tool for PraisonAI Agents.

Send messages via WhatsApp Business API.

Usage:
    from praisonai_tools import WhatsAppTool
    
    wa = WhatsAppTool()
    wa.send_message(to="+1234567890", message="Hello!")

Environment Variables:
    WHATSAPP_ACCESS_TOKEN: WhatsApp Business API access token
    WHATSAPP_PHONE_NUMBER_ID: Phone number ID
"""

import os
import logging
from typing import Any, Dict, Optional, Union

from praisonai_tools.tools.base import BaseTool

logger = logging.getLogger(__name__)


class WhatsAppTool(BaseTool):
    """Tool for WhatsApp messaging."""
    
    name = "whatsapp"
    description = "Send messages via WhatsApp Business API."
    
    def __init__(
        self,
        access_token: Optional[str] = None,
        phone_number_id: Optional[str] = None,
    ):
        self.access_token = access_token or os.getenv("WHATSAPP_ACCESS_TOKEN")
        self.phone_number_id = phone_number_id or os.getenv("WHATSAPP_PHONE_NUMBER_ID")
        super().__init__()
    
    def run(
        self,
        action: str = "send",
        to: Optional[str] = None,
        message: Optional[str] = None,
        **kwargs
    ) -> Union[str, Dict[str, Any]]:
        action = action.lower().replace("-", "_")
        
        if action == "send":
            return self.send_message(to=to, message=message)
        elif action == "send_template":
            return self.send_template(to=to, template=kwargs.get("template"))
        else:
            return {"error": f"Unknown action: {action}"}
    
    def send_message(self, to: str, message: str) -> Dict[str, Any]:
        """Send text message."""
        if not to or not message:
            return {"error": "to and message are required"}
        if not self.access_token or not self.phone_number_id:
            return {"error": "WHATSAPP_ACCESS_TOKEN and WHATSAPP_PHONE_NUMBER_ID required"}
        
        try:
            import requests
        except ImportError:
            return {"error": "requests not installed"}
        
        try:
            url = f"https://graph.facebook.com/v18.0/{self.phone_number_id}/messages"
            headers = {
                "Authorization": f"Bearer {self.access_token}",
                "Content-Type": "application/json",
            }
            data = {
                "messaging_product": "whatsapp",
                "to": to,
                "type": "text",
                "text": {"body": message},
            }
            resp = requests.post(url, headers=headers, json=data, timeout=10)
            result = resp.json()
            
            if "error" in result:
                return {"error": result["error"].get("message", str(result["error"]))}
            return {"success": True, "message_id": result.get("messages", [{}])[0].get("id")}
        except Exception as e:
            logger.error(f"WhatsApp send error: {e}")
            return {"error": str(e)}
    
    def send_template(self, to: str, template: str, language: str = "en_US") -> Dict[str, Any]:
        """Send template message."""
        if not to or not template:
            return {"error": "to and template are required"}
        if not self.access_token or not self.phone_number_id:
            return {"error": "WHATSAPP_ACCESS_TOKEN and WHATSAPP_PHONE_NUMBER_ID required"}
        
        try:
            import requests
        except ImportError:
            return {"error": "requests not installed"}
        
        try:
            url = f"https://graph.facebook.com/v18.0/{self.phone_number_id}/messages"
            headers = {
                "Authorization": f"Bearer {self.access_token}",
                "Content-Type": "application/json",
            }
            data = {
                "messaging_product": "whatsapp",
                "to": to,
                "type": "template",
                "template": {"name": template, "language": {"code": language}},
            }
            resp = requests.post(url, headers=headers, json=data, timeout=10)
            result = resp.json()
            
            if "error" in result:
                return {"error": result["error"].get("message", str(result["error"]))}
            return {"success": True, "message_id": result.get("messages", [{}])[0].get("id")}
        except Exception as e:
            logger.error(f"WhatsApp template error: {e}")
            return {"error": str(e)}


def send_whatsapp_message(to: str, message: str) -> Dict[str, Any]:
    """Send WhatsApp message."""
    return WhatsAppTool().send_message(to=to, message=message)
